import abc
from dataclasses import dataclass, field
from datetime import timedelta
from enum import Enum
from functools import lru_cache
from math import ceil
import numpy as np
from typing import List, Optional, Tuple, Any, Union

import toloka.client as toloka

from . import base, mapping


@dataclass
class PricingStrategy(abc.ABC):
    ...


@dataclass
class StaticPricingStrategy(PricingStrategy):
    ...


@dataclass
class DynamicPricingStrategy(PricingStrategy):
    max_ratio: float  # max price for "good" work divided by price for "ok" work


@dataclass
class PoolPricingConfig:
    assignment_price: float
    real_tasks_count: int
    control_tasks_count: int

    def get_tasks_count(self) -> int:
        return self.real_tasks_count + self.control_tasks_count


@dataclass
class TrainingConfig:
    training_tasks_in_task_suite_count: int
    task_suites_required_to_pass: int
    max_assignment_duration: timedelta

    def get_task_duration_hint(self) -> timedelta:
        return self.max_assignment_duration / self.training_tasks_in_task_suite_count / training_overhead_multiplier

    @staticmethod
    def from_toloka_training(
        training: toloka.Training,
        task_duration_hint: Optional[timedelta] = None,
        available_task_count: Optional[int] = None,
    ) -> 'TrainingConfig':
        if task_duration_hint is None:
            return TrainingConfig(
                training.training_tasks_in_task_suite_count,
                training.task_suites_required_to_pass,
                timedelta(seconds=training.assignment_max_duration_seconds),
            )
        assert available_task_count is not None
        training_time = (
            timedelta(seconds=training.assignment_max_duration_seconds)
            * training.task_suites_required_to_pass
            / training_overhead_multiplier
        )
        return choose_default_training_option(
            get_training_options(task_duration_hint, available_task_count, training_time)
        )


toloka_avg_price_per_hour = {'EN': 1.0, 'RU': 0.6, 'TR': 1.2, 'KK': 0.8}

toloka_commission = 0.3
min_toloka_commission = 0.001

min_assignment_cost = 0.005
threshold_for_assignment_cost_granularity = 1.0

min_assignment_duration = timedelta(seconds=30)
max_assignment_duration = timedelta(minutes=20)

min_tasks_in_assignment = 1
max_tasks_in_assignment = 30

min_training_assignments_to_pass = 1
max_training_assignments_to_pass = 5

min_training_tasks_to_pass = 2
max_training_tasks_to_pass = 30

assert max_tasks_in_assignment >= min_tasks_in_assignment
assert max_training_assignments_to_pass >= min_training_assignments_to_pass
assert max_training_tasks_to_pass >= min_training_tasks_to_pass

assert min_training_tasks_to_pass >= min_tasks_in_assignment * min_training_assignments_to_pass
assert max_training_tasks_to_pass <= max_tasks_in_assignment * max_training_assignments_to_pass

default_training_time = timedelta(minutes=3)

# configures extra time to account for not accustomed to the task workers
# do not change it between launches with same data, because internal calculations depend on it
training_overhead_multiplier = 10


@lru_cache(maxsize=None)
def _choose(n: int, k: int) -> int:
    result = 1
    for j in range(k):
        result *= n - j
        result //= j + 1
    return result


@lru_cache(maxsize=None)
def _get_random_chance(
    control_task_count: int,
    correct_control_task_count_for_acceptance: int,
    num_classes: int,
) -> float:
    sum_probability = 0.0
    for applicable_control_task_count in range(correct_control_task_count_for_acceptance, control_task_count + 1):
        multiplier = _choose(control_task_count, applicable_control_task_count)
        random_chance = (
            (1 / num_classes) ** applicable_control_task_count
            * (1 - 1 / num_classes) ** (control_task_count - applicable_control_task_count)
            * multiplier
        )
        sum_probability += random_chance
    return min(1.0, max(0.0, sum_probability))


def _get_num_classes(task_mapping: mapping.TaskMapping) -> int:
    return len(task_mapping.output_mapping[0].obj_meta.type.possible_values())


@dataclass
class PoolPricingConfigWithCalculatedProperties:
    config: PoolPricingConfig
    price_per_hour: float
    robustness: float
    control_tasks_ratio: float
    assignment_duration: timedelta


def _produce_pricing_config_with_calculated_properties(
    task_duration_hint: timedelta,
    random_chance: float,
    config: PoolPricingConfig,
) -> PoolPricingConfigWithCalculatedProperties:
    tasks_count = config.get_tasks_count()
    assignment_per_hour = timedelta(hours=1).total_seconds() / (task_duration_hint * tasks_count).total_seconds()
    return PoolPricingConfigWithCalculatedProperties(
        config=config,
        price_per_hour=config.assignment_price * assignment_per_hour,
        robustness=1 - random_chance,
        control_tasks_ratio=config.control_tasks_count / tasks_count,
        assignment_duration=task_duration_hint * tasks_count,
    )


def _readable_duration(duration: timedelta) -> str:
    # round up to seconds
    return str(timedelta(seconds=ceil(duration.total_seconds())))


def _get_price_per_hour(lang: str) -> float:
    return toloka_avg_price_per_hour.get(lang, toloka_avg_price_per_hour[base.DEFAULT_LANG])


def get_control_tasks_for_accept_from_ratio(ratio: float, control_tasks_count: int) -> int:
    return ceil(ratio * control_tasks_count)


def calculate_properties_for_pricing_config(
    config: PoolPricingConfig,
    task_duration_hint: timedelta,
    correct_control_task_ratio_for_acceptance: float,
    task_mapping: mapping.TaskMapping,
) -> PoolPricingConfigWithCalculatedProperties:
    num_classes = _get_num_classes(task_mapping)
    random_chance = _get_random_chance(
        config.control_tasks_count,
        get_control_tasks_for_accept_from_ratio(correct_control_task_ratio_for_acceptance, config.control_tasks_count),
        num_classes,
    )
    return _produce_pricing_config_with_calculated_properties(task_duration_hint, random_chance, config)


def round_price(value: float) -> float:
    # need for round:
    # 1. to remove displays such as '0.0299999999999' instead of '0.3', which can be got from float price calculations
    # 2. we have different granularity for prices lower and higher than 1$ threshold
    digits = 2 if value > threshold_for_assignment_cost_granularity else 3
    return round(value, digits)


def _get_assignment_params(
    task_count: int,
    task_duration_hint: timedelta,
    price_per_hour: Optional[float],
    fixed_price: Optional[float] = None,
) -> Tuple[timedelta, float, float]:
    price_per_second = price_per_hour / (60 * 60) if price_per_hour is not None else None

    duration = task_count * task_duration_hint

    price = max(
        min_assignment_cost,
        round_price(duration.total_seconds() * price_per_second) if fixed_price is None else fixed_price,
    )
    commission = toloka_commission * price

    return duration, price, commission


def get_pricing_options(
    task_duration_hint: timedelta,
    pricing_strategy: PricingStrategy,
    lang: str,
    task_mapping: mapping.TaskMapping,
    max_random_chance_for_control_tasks: float = 1 / 3,
    max_control_task_percentage_on_page: float = 1 / 3,
    correct_control_task_ratio_for_acceptance: float = 1.0,
    plot: bool = False,
) -> List[PoolPricingConfigWithCalculatedProperties]:
    assert len(task_mapping.output_mapping) == 1
    assert issubclass(task_mapping.output_mapping[0].obj_meta.type, base.Label)
    num_classes = _get_num_classes(task_mapping)

    if not isinstance(task_duration_hint, timedelta):
        raise NotImplementedError(task_duration_hint)

    if not isinstance(pricing_strategy, StaticPricingStrategy):
        raise NotImplementedError(pricing_strategy)

    suitable_options = []

    # min_control_task_count = ceil(log(max_random_chance_for_control_tasks) / log(1 / num_classes))

    price_per_hour = _get_price_per_hour(lang)

    for task_count in range(min_tasks_in_assignment, max_tasks_in_assignment):
        assignment_duration, assignment_price, assignment_commission = _get_assignment_params(
            task_count, task_duration_hint, price_per_hour
        )

        for control_task_count in range(1, int(ceil(max_control_task_percentage_on_page * task_count)) + 1):
            random_chance = _get_random_chance(
                control_task_count,
                get_control_tasks_for_accept_from_ratio(correct_control_task_ratio_for_acceptance, control_task_count),
                num_classes,
            )
            suitable = (
                random_chance <= max_random_chance_for_control_tasks
                and min_assignment_duration <= assignment_duration <= max_assignment_duration
                and assignment_commission >= min_toloka_commission
            )

            if suitable:
                suitable_options.append(
                    _produce_pricing_config_with_calculated_properties(
                        task_duration_hint,
                        random_chance,
                        PoolPricingConfig(
                            real_tasks_count=task_count - control_task_count,
                            control_tasks_count=control_task_count,
                            assignment_price=assignment_price,
                        ),
                    )
                )

    if plot:
        plot_options(
            options=suitable_options,
            task_duration_hint=task_duration_hint,
            max_random_chance_for_control_tasks=max_random_chance_for_control_tasks,
            max_control_task_percentage_on_page=max_control_task_percentage_on_page,
            correct_control_task_ratio_for_acceptance=correct_control_task_ratio_for_acceptance,
            default_option=choose_default_option(suitable_options),
            num_classes=num_classes,
            lang=lang,
            price_per_hour=price_per_hour,
            html_path='pricing.html',
        )

    return suitable_options


def get_annotation_pricing_options(
    task_duration_hint: timedelta,
    pricing_strategy: PricingStrategy,
    lang: str,
    task_mapping: mapping.TaskMapping,
) -> List[PoolPricingConfig]:
    if not isinstance(pricing_strategy, StaticPricingStrategy):
        raise NotImplementedError(pricing_strategy)

    suitable_options = []

    price_per_hour = _get_price_per_hour(lang)

    for task_count in range(min_tasks_in_assignment, max_tasks_in_assignment):
        assignment_duration, assignment_price, assignment_commission = _get_assignment_params(
            task_count, task_duration_hint, price_per_hour
        )

        suitable = (
            min_assignment_duration <= assignment_duration <= max_assignment_duration
            and assignment_commission >= min_toloka_commission
        )

        if suitable:
            suitable_options.append(
                PoolPricingConfig(real_tasks_count=task_count, control_tasks_count=0, assignment_price=assignment_price)
            )
    return suitable_options


def choose_default_annotation_option(options: List[PoolPricingConfig]) -> PoolPricingConfig:
    assert options, 'No available options'
    return options[len(options) // 2]


def get_expert_pricing_options(
    task_duration_hint: timedelta,
    task_mapping: mapping.TaskMapping,  # todo we may use this in the future
    avg_price_per_hour: Optional[float] = None,
) -> List[PoolPricingConfig]:
    suitable_options = []

    # todo плата экспертов должна быть выше?
    minimal_pay_for_inhouse = avg_price_per_hour is None

    for task_count in range(min_tasks_in_assignment, max_tasks_in_assignment):
        assignment_duration, assignment_price, assignment_commission = _get_assignment_params(
            task_count,
            task_duration_hint,
            avg_price_per_hour,
            fixed_price=min_assignment_cost if minimal_pay_for_inhouse else None,
        )

        suitable = min_assignment_duration <= assignment_duration <= max_assignment_duration and (
            minimal_pay_for_inhouse or assignment_commission >= min_toloka_commission
        )

        if suitable:
            suitable_options.append(
                PoolPricingConfig(real_tasks_count=task_count, control_tasks_count=0, assignment_price=assignment_price)
            )

    return suitable_options


def get_training_options(
    task_duration_hint: timedelta,
    available_task_count: int,
    training_time: timedelta = default_training_time,
) -> List[TrainingConfig]:
    task_count = ceil(training_time / task_duration_hint)
    assert min_training_tasks_to_pass <= task_count <= max_training_tasks_to_pass, (
        'You are trying to create either too short or too long training. Expected from '
        f'{min_training_tasks_to_pass} to {max_training_tasks_to_pass} tasks, got {task_count}'
    )
    assert task_count <= available_task_count, (
        'You do not have enough tasks for this training. Expected at least '
        f'{task_count} tasks for this parameters, got {available_task_count}'
    )
    configs = []
    for assignments_count in range(min_training_assignments_to_pass, max_training_assignments_to_pass + 1):
        tasks_in_assignment = ceil(task_count / assignments_count)
        if min_tasks_in_assignment <= tasks_in_assignment <= max_tasks_in_assignment:
            config = TrainingConfig(
                training_tasks_in_task_suite_count=tasks_in_assignment,
                task_suites_required_to_pass=assignments_count,
                max_assignment_duration=tasks_in_assignment * task_duration_hint * training_overhead_multiplier,
            )
            configs.append(config)

    return configs


def choose_default_training_option(options: List[TrainingConfig]) -> TrainingConfig:
    assert options, 'No available options'
    return options[len(options) // 2]


def choose_default_option(
    suitable_options: List[PoolPricingConfigWithCalculatedProperties],
) -> PoolPricingConfigWithCalculatedProperties:
    assert suitable_options, 'No suitable options found'

    filtered_options = list(suitable_options)

    # choose option by average over chosen dimensions
    for value_f in (
        lambda choice: choice.robustness,
        lambda choice: choice.config.control_tasks_count,
        lambda choice: choice.config.control_tasks_count + choice.config.real_tasks_count,
    ):
        values = np.array([value_f(option) for option in filtered_options])
        avg_value = values[np.abs(values - values.mean()).argmin()]
        filtered_options = [option for option in filtered_options if value_f(option) == avg_value]

    return filtered_options[0]  # there can be multiple options with same last filter value


def choose_default_expert_option(
    suitable_options: List[PoolPricingConfig],
    avg_price_per_hour: Optional[float],
) -> PoolPricingConfig:
    assert suitable_options, 'No suitable options found'
    return (min if avg_price_per_hour is not None else max)(
        suitable_options, key=lambda config: config.get_tasks_count()
    )


def _get_plots_arguments(
    options: List[PoolPricingConfigWithCalculatedProperties],
) -> Tuple[List[float], List[int], List[float], List[int], List[List[Union[float, int]]]]:
    x, y, r, t, texts = [], [], [], [], []

    for option in options:
        tasks_count = option.config.get_tasks_count()
        task_price_cents = (option.config.assignment_price * 100) / option.config.real_tasks_count

        x.append(option.robustness)
        y.append(int(option.assignment_duration.total_seconds()))
        r.append(option.control_tasks_ratio)
        t.append(tasks_count)

        texts.append(
            [
                task_price_cents,
                option.config.assignment_price,
                option.robustness,
                option.control_tasks_ratio,
                option.config.real_tasks_count,
                option.config.control_tasks_count,
                _readable_duration(option.assignment_duration),
            ]
        )
    return x, y, r, t, texts


TEMPLATE = (
    'Price per task, ¢: %{customdata[0]:.2f}<br>'
    'Price per assignment, $: %{customdata[1]:.2f}<br>'
    'Robustness: %{customdata[2]:.2f}<br>'
    'Control task ratio: %{customdata[3]:.2f}<br>'
    'Real task count: %{customdata[4]}<br>'
    'Control task count: %{customdata[5]}<br>'
    'Assignment duration: %{customdata[6]}<extra></extra>'
)


# todo marker & font size


class PlotDimensions(Enum):
    DURATION = 'Assignment duration (seconds)'
    ROBUSTNESS = 'Random robustness'
    CONTROL_TASK_RATIO = 'Control task ratio'


class AvailableConfigurations(Enum):
    ROBUSTNESS_AND_DURATION = (PlotDimensions.ROBUSTNESS, PlotDimensions.DURATION, PlotDimensions.CONTROL_TASK_RATIO)
    CONTROL_AND_DURATION = (PlotDimensions.CONTROL_TASK_RATIO, PlotDimensions.DURATION, PlotDimensions.ROBUSTNESS)
    ROBUSTNESS_AND_CONTROL = (PlotDimensions.ROBUSTNESS, PlotDimensions.CONTROL_TASK_RATIO, PlotDimensions.DURATION)


@dataclass
class PlotConfiguration:
    dimensions: AvailableConfigurations
    coloraxis: str = field(init=False)
    x_index: int = field(init=False)
    y_index: int = field(init=False)
    c_index: int = field(init=False)
    reversescale: bool = field(init=False)
    need_x_adjustment: bool = field(init=False)
    need_y_adjustment: bool = field(init=False)

    def __post_init__(self):
        if self.dimensions == AvailableConfigurations.ROBUSTNESS_AND_DURATION:
            x, y, c, axis = 0, 1, 2, 1
        elif self.dimensions == AvailableConfigurations.CONTROL_AND_DURATION:
            x, y, c, axis = 2, 1, 0, 2
        elif self.dimensions == AvailableConfigurations.ROBUSTNESS_AND_CONTROL:
            x, y, c, axis = 0, 2, 1, 3
        else:
            raise ValueError(f'unsupported dimensions: {self.dimensions}')
        self.x_index, self.y_index, self.c_index = x, y, c
        self.coloraxis = f'coloraxis{axis}'
        self.reversescale = self.dimensions != AvailableConfigurations.CONTROL_AND_DURATION
        self.need_x_adjustment = self.dimensions != AvailableConfigurations.CONTROL_AND_DURATION
        self.need_y_adjustment = self.dimensions != AvailableConfigurations.ROBUSTNESS_AND_CONTROL


def _scatter_plot(
    configuration: PlotConfiguration,
    options: List[PoolPricingConfigWithCalculatedProperties],
    task_duration_hint: timedelta,
    max_random_chance_for_control_tasks: float,
    default_option: Optional[PoolPricingConfig] = None,
    row: Optional[int] = None,
    col: Optional[int] = None,
    html_path: Optional[str] = None,
    showfig: bool = True,
    fig: Any = None,  # todo - чтобы написать тип plotly.graph_objs.Figure, надо сначала заимпортить plotly
):
    import plotly.graph_objs as go

    if default_option is None:
        default_option = choose_default_option(options)
    params = _get_plots_arguments(options)
    x, _, _, t, _ = params

    d_params = _get_plots_arguments([default_option])
    if fig is None:
        fig = go.Figure()

    plot_x, plot_y = params[configuration.x_index], params[configuration.y_index]
    plot_dx, plot_dy = d_params[configuration.x_index], d_params[configuration.y_index]
    color, dcolor = params[configuration.c_index], d_params[configuration.c_index]
    texts, dtexts = params[-1], d_params[-1]

    fig.add_trace(
        go.Scatter(
            x=plot_x,
            y=plot_y,
            customdata=texts,
            marker=dict(
                coloraxis=configuration.coloraxis, color=color, size=10, line=dict(width=1, color='DarkSlateGrey')
            ),
            mode='markers',
            name='choices',
            hovertemplate=TEMPLATE,
        ),
        row=row,
        col=col,
    )
    fig.add_trace(
        go.Scatter(
            x=plot_dx,
            y=plot_dy,
            customdata=dtexts,
            marker=dict(coloraxis=configuration.coloraxis, color=dcolor, size=12, line=dict(width=2, color='black')),
            mode='markers',
            name='Default choice',
            hovertemplate=f'<b>Recommended option</b><br>{TEMPLATE}',
        ),
        row=row,
        col=col,
    )

    if configuration.need_x_adjustment:
        fig.add_vline(
            x=1.0 - max_random_chance_for_control_tasks,
            line_dash='dashdot',
            annotation_text='min robustness',
            line_color='red',
            line_width=1,
            annotation=dict(textangle=90, y=0.7, font_color='darkred', font_size=8),
            row=row,
            col=col,
        )
        fig.update_xaxes(range=[1.0 - max_random_chance_for_control_tasks - 0.01, max(x) + 0.01], row=row, col=col)

    if configuration.need_y_adjustment:
        duration_thresholds = [
            (max_assignment_duration / 4, 'green', 'comfort'),
            (max_assignment_duration * 3 / 4, 'orange', 'inconvenient'),
            (None, 'red', 'critical'),
        ]

        for i in range(len(duration_thresholds)):
            y1, color, name = duration_thresholds[i]

            if i > 0:
                y0, _, _ = duration_thresholds[i - 1]
            else:
                y0 = min_assignment_duration

            if i == len(duration_thresholds) - 1:
                y1 = max_assignment_duration

            fig.add_hrect(
                y0=int(y0.total_seconds()),
                y1=int(y1.total_seconds()),
                fillcolor=color,
                opacity=0.1,
                annotation_position='bottom right',
                annotation_text=f'{name} duration',
                annotation=dict(font_color=f'dark{color}', font_size=8),
                row=row,
                col=col,
            )

        task_duration_hint_seconds = int(task_duration_hint.total_seconds())

        fig.update_yaxes(
            range=[min(t) * task_duration_hint_seconds - 50, max(t) * task_duration_hint_seconds + 50], row=row, col=col
        )

    if row is None and col is None:
        fig.update_layout(
            margin=dict(l=0, r=0, t=0, b=0),
            legend=dict(x=0.5, xanchor='center'),
            legend_orientation='h',
            xaxis_title=configuration.dimensions.value[0].value,
            yaxis_title=configuration.dimensions.value[1].value,
            showlegend=False,
            **{
                configuration.coloraxis: dict(
                    colorscale='RdYlGn',
                    colorbar=dict(title=configuration.dimensions.value[2].value),
                    reversescale=configuration.reversescale,
                )
            },
        )

    if html_path is not None:
        fig.write_html(html_path)

    if showfig:
        fig.show()


def plot_options(
    options: List[PoolPricingConfigWithCalculatedProperties],
    task_duration_hint: timedelta,
    max_random_chance_for_control_tasks: float,
    correct_control_task_ratio_for_acceptance: float,
    num_classes: int,
    lang: str,
    price_per_hour: float,
    max_control_task_percentage_on_page: float,
    default_option: Optional[PoolPricingConfigWithCalculatedProperties] = None,
    html_path: Optional[str] = None,
    showfig: bool = True,
):
    from plotly.subplots import make_subplots

    fig = make_subplots(
        rows=2,
        cols=2,
        vertical_spacing=0.2,
        horizontal_spacing=0.2,
        subplot_titles=('', '', '', 'Params'),
    )

    configs = (
        (AvailableConfigurations.ROBUSTNESS_AND_DURATION, 1, 1),
        (AvailableConfigurations.CONTROL_AND_DURATION, 1, 2),
        (AvailableConfigurations.ROBUSTNESS_AND_CONTROL, 2, 1),
    )

    for config, row, col in configs:
        _scatter_plot(
            PlotConfiguration(config),
            options,
            task_duration_hint,
            max_random_chance_for_control_tasks,
            default_option,
            row=row,
            col=col,
            showfig=False,
            fig=fig,
        )

    params = (
        f'Task duration hint: {_readable_duration(task_duration_hint)}<br>'
        f'"{lang}" avg price per hour: {price_per_hour:.2f}$<br>'
        f'Number of classes: {num_classes}<br>'
        f'Control correct ratio for accept: {correct_control_task_ratio_for_acceptance:.2f}<br>'
        f'Max random chance: {max_random_chance_for_control_tasks:.2f}<br>'
        f'Max control task ratio: {max_control_task_percentage_on_page:.2f}<br>'
        f'Number of options: {len(options)}'
    )

    fig.add_annotation(
        text=params,
        align='left',
        showarrow=False,
        x=1.1,
        y=0.8,
        bordercolor='black',
        borderwidth=1,
        bgcolor='white',
        font_size=25,
        row=2,
        col=2,
    )

    fig.update_xaxes(showgrid=False, showticklabels=False, row=2, col=2)
    fig.update_yaxes(showgrid=False, showticklabels=False, row=2, col=2)

    fig.update_layout(
        legend=dict(x=0.5, xanchor='center'),
        legend_orientation='h',
        title='Pricing',
        title_x=0.5,
        coloraxis1=dict(
            colorscale='RdYlGn',
            colorbar=dict(title=PlotDimensions.CONTROL_TASK_RATIO.value, x=0.46, y=0.8, len=0.4),
            reversescale=True,
            showscale=True,
        ),
        coloraxis2=dict(
            colorscale='RdYlGn',
            colorbar=dict(title=PlotDimensions.ROBUSTNESS.value, x=1.05, y=0.8, len=0.4),
            showscale=True,
        ),
        coloraxis3=dict(
            colorscale='RdYlGn',
            colorbar=dict(title=PlotDimensions.DURATION.value, x=0.46, y=0.2, len=0.4),
            reversescale=True,
            showscale=True,
        ),
        showlegend=False,
        height=1000,
        width=1600,
    )

    for i, (config, row, col) in enumerate(configs):
        x_dim, y_dim, _ = config.value
        fig['layout'][f'xaxis{i + 1}']['title'] = x_dim.value
        fig['layout'][f'yaxis{i + 1}']['title'] = y_dim.value

    if html_path is not None:
        fig.write_html(html_path)

    if showfig:
        fig.show()



import abc
import datetime
from collections import defaultdict
from dataclasses import dataclass
import logging
from random import shuffle
from typing import List, Dict, Optional, Tuple, Iterable, Set

import toloka.client as toloka

from .. import base, classification, control, duration, mapping
from ..worker import Human, Model, Worker

logger = logging.getLogger(__name__)


@dataclass
class SolutionEvaluation:
    ok: bool
    confidence: float

    worker_labels: List[classification.WorkerLabel]


def collect_evaluations_from_check_assignments(
    assignments: List[toloka.Assignment],
    check_task_mapping: mapping.TaskMapping,
    pool_input_objects: Optional[List[mapping.Objects]],
    aggregation_algorithm: classification.AggregationAlgorithm,
    confidence_threshold: float,
    worker_weights: Optional[Dict[str, float]] = None,
) -> Dict[mapping.TaskID, SolutionEvaluation]:
    if pool_input_objects is None:
        # TODO: questionable
        # In feedback loop, we don't know exactly, which markup solutions we've checked in case of check sample.
        # So pool input objects are generated from check assignments. Order of solutions in relation to input objects
        # will be lost, but it doesn't matter for feedback loop, solutions will be reordered according to confidence.
        # In other cases, i.e. classification tasks, we want to preserve order of solutions in relation to input
        # objects, and it will be preserved, because we will have pool input objects.
        pool_input_objects = []
        task_ids = set()
        for assignment in assignments:
            for task_id, input_objects, _ in mapping.iterate_assignment(assignment, check_task_mapping):
                if task_id in task_ids:
                    continue
                task_ids.add(task_id)
                pool_input_objects.append(input_objects)

    evaluations = []
    for task_labels_probas, worker_labels in classification.collect_labels_probas_from_assignments(
        assignments,
        check_task_mapping,
        pool_input_objects,
        aggregation_algorithm,
        worker_weights,
    ):
        # TODO: support other evaluation types
        # TODO: add individual workers evaluations like in base loop
        # TODO: may be None
        assert task_labels_probas is not None
        confidence = task_labels_probas[base.BinaryEvaluation(ok=True)]
        assert worker_labels
        evaluations.append(
            SolutionEvaluation(
                ok=confidence >= confidence_threshold, confidence=confidence, worker_labels=worker_labels
            )
        )

    return {
        check_task_mapping.task_id(input_objects): evaluation
        for input_objects, evaluation in zip(pool_input_objects, evaluations)
    }


@dataclass
class AssignmentCheckSample:
    max_tasks_to_check: Optional[int]
    assignment_accuracy_finalization_threshold: Optional[float]


def find_markup_solutions_to_check(
    markup_assignments: List[mapping.AssignmentSolutions],
    checked_solutions: Set[mapping.TaskID],
    check_task_mapping: mapping.TaskMapping,
    check_sample: Optional[AssignmentCheckSample],
) -> Tuple[List[Tuple[mapping.Objects, Worker]], Set[mapping.TaskID]]:
    solutions_to_check = []
    already_checked_solutions = set()
    for assignment, solutions in markup_assignments:
        assert assignment.status == toloka.Assignment.SUBMITTED
        if assignment.id is not None:
            worker = Human(assignment)
        else:
            worker = Model(assignment.user_id, None)
        not_checked_solutions = []
        for input_objects, output_objects in solutions:
            solution = input_objects + output_objects
            solution_id = check_task_mapping.task_id(solution)
            if solution_id in checked_solutions:
                already_checked_solutions.add(solution_id)
            else:
                not_checked_solutions.append(solution)
        from_model = not assignment.id
        solutions_to_check += [
            (assignment_solution, worker)
            for assignment_solution in sample_solutions_to_check(not_checked_solutions, check_sample, from_model)
        ]
    return solutions_to_check, already_checked_solutions


def sample_solutions_to_check(
    solutions: List[mapping.Objects],
    check_sample: Optional[AssignmentCheckSample],
    from_model: bool = False,
) -> List[mapping.Objects]:
    if check_sample is None or check_sample.max_tasks_to_check is None or from_model:
        # check sample is disabled, or it is model solutions from single assignment with arbitrary task count
        return solutions
    shuffle(solutions)
    return solutions[: check_sample.max_tasks_to_check]


def get_tasks_attempts(
    assignments: Iterable[mapping.AssignmentSolutions],
    markup_task_mapping: mapping.TaskMapping,
    with_model: bool = True,
) -> Dict[mapping.TaskID, int]:
    task_id_to_attempts = defaultdict(int)
    for assignment, solutions in assignments:
        for input_objects, _ in solutions:
            if not assignment.id and not with_model:
                increase = 0
            else:
                increase = 1
            task_id_to_attempts[markup_task_mapping.task_id(input_objects)] += increase
    return task_id_to_attempts


def process_assignment_by_duration(
    reject_rules: List[control.Rule],
    block_rules: List[control.Rule],
    assignment: toloka.Assignment,
    lang: str,
    client: toloka.TolokaClient,
    assignment_duration_hint: datetime.timedelta,
) -> Optional[toloka.Assignment.Status]:
    assignment_duration = assignment.submitted - assignment.created
    verdict: Optional[toloka.Assignment.Status] = None
    # We need to first block user, if needed, and only after that set the assignment status:
    # in feedback loop case, both SUBMITTED and REJECTED assignments can arrive here.
    # We do not want to re-apply block to users from already rejected assignments,
    # we do it by taking assignment.status into account in BlockUSer.
    # However, if we set assignment statuses to rejected first, and deal with blocks later,
    # and feedback loop is interrupted between these steps, after the restart user wouldn't be blocked
    # because the assignment will be deemed as "already processed" by BlockUser

    for rule in block_rules:
        restriction = rule.apply(
            value=assignment_duration,
            assignment_duration_hint=assignment_duration_hint,
            client=client,
            user_id=assignment.user_id,
            pool_id=assignment.pool_id,
            block_start=assignment.submitted,
            assignment_status=assignment.status,
        )
        if restriction is not None:
            logger.debug(f'add restriction {restriction} for user {assignment.user_id} by {rule.predicate}')

    for rule in reject_rules:
        verdict = rule.apply(
            value=assignment_duration,
            assignment_duration_hint=assignment_duration_hint,
            client=client,
            assignment=assignment,
            public_comment=assignment_short_rejection_comment[lang],
        )
        if verdict is not None:
            logger.debug(f'set assignment {assignment.id} to status {verdict.value} by {rule.predicate}')
            break
    return verdict


# Some prior assignments checks, i.e. check for completion time, can be performed before real evaluation,
# to avoid possibly time-cost consuming (in case of feedback loop) evaluation.
def prior_filter_assignments(
    client: toloka.TolokaClient,
    submitted_assignments: Iterable[mapping.AssignmentSolutions],
    control_params: control.Control,
    lang: str,
    task_duration_function: duration.TaskDurationFunction,
) -> Tuple[List[mapping.AssignmentSolutions], List[mapping.AssignmentSolutions]]:
    reject_rules = control_params.filter_rules(
        predicate_type=control.AssignmentDurationPredicate, action_type=control.SetAssignmentStatus
    )

    block_rules = control_params.filter_rules(
        predicate_type=control.AssignmentDurationPredicate, action_type=control.BlockUser
    )

    filtered_assignments = []
    fast_assignments = []
    for assignment_solution in submitted_assignments:
        assignment, solutions = assignment_solution
        assignment_duration_hint = sum(
            [task_duration_function(input_objects) for (input_objects, _) in solutions],
            start=datetime.timedelta(seconds=0),
        )
        verdict = process_assignment_by_duration(
            reject_rules=reject_rules,
            block_rules=block_rules,
            assignment=assignment,
            lang=lang,
            client=client,
            assignment_duration_hint=assignment_duration_hint,
        )
        if verdict is None:
            filtered_assignments.append(assignment_solution)
        elif verdict == toloka.Assignment.Status.REJECTED:
            fast_assignments.append(assignment_solution)
        else:
            assert False, f'Found fast assignment {assignment.id} with unexpected status {assignment.status}'
    return filtered_assignments, fast_assignments


@dataclass
class AssignmentEvaluation:
    assignment: toloka.Assignment
    ok_checks: int
    total_checks: int
    objects: int
    incorrect_solution_indexes: List[int]

    def get_accuracy(self) -> Optional[float]:
        if self.total_checks == 0:
            return None
        return self.ok_checks / self.total_checks

    def get_evaluation_recall(self) -> float:
        return self.total_checks / self.objects


class AssignmentAccuracyEvaluationStrategy:
    can_have_zero_checks = False

    @abc.abstractmethod
    def ok(self, task: toloka.Task, solution: mapping.TaskSingleSolution) -> Optional[bool]:
        ...

    def evaluate_assignment(self, assignment_solutions: mapping.AssignmentSolutions) -> AssignmentEvaluation:
        ok_checks, total_checks = 0, 0
        incorrect_solutions_indexes = []
        assignment, solutions = assignment_solutions
        for i, (task, solution) in enumerate(zip(assignment.tasks, solutions)):
            ok = self.ok(task, solution)
            if ok is None:
                continue
            total_checks += 1
            if ok:
                ok_checks += 1
            else:
                incorrect_solutions_indexes.append(i)
        assert self.can_have_zero_checks or total_checks > 0, f'zero checks for assignment {assignment}'

        return AssignmentEvaluation(
            assignment=assignment,
            ok_checks=ok_checks,
            total_checks=total_checks,
            objects=len(solutions),
            incorrect_solution_indexes=incorrect_solutions_indexes,
        )


# driven by evaluations in check pool
@dataclass
class CheckAssignmentAccuracyEvaluationStrategy(AssignmentAccuracyEvaluationStrategy):
    solution_id_to_evaluation: Dict[mapping.TaskID, SolutionEvaluation]
    check_task_mapping: mapping.TaskMapping

    def ok(self, task: toloka.Task, solution: mapping.TaskSingleSolution) -> Optional[bool]:
        input_objects, output_objects = solution
        solution_id = self.check_task_mapping.task_id(input_objects + output_objects)
        if solution_id not in self.solution_id_to_evaluation:
            return None
        return self.solution_id_to_evaluation[solution_id].ok


# driven by control tasks in each assignment
@dataclass
class ControlTasksAssignmentAccuracyEvaluationStrategy(AssignmentAccuracyEvaluationStrategy):
    task_mapping: mapping.TaskMapping
    can_have_zero_checks: bool = False

    def ok(self, task: toloka.Task, solution: mapping.TaskSingleSolution) -> Optional[bool]:
        # we count on embedding control task single expected solution to Task.known_solutions
        if not task.known_solutions:
            return None
        assert len(task.known_solutions) == 1
        expected_solution = task.known_solutions[0].output_values
        _, output_objects = solution
        actual_solution = self.task_mapping.to_solution(output_objects).output_values

        # we check only values which are in expected solution, some output values,
        # i.e. comments, is meaninglessly to check
        return {**actual_solution, **expected_solution} == actual_solution


class CustomEvaluationStrategy(AssignmentAccuracyEvaluationStrategy):
    @abc.abstractmethod
    def update(self, *args, **kwargs):
        ...


def apply_rules_to_assignment(
    set_verdict_rules: List[control.Rule],
    block_rules: List[control.Rule],
    evaluation: AssignmentEvaluation,
    client: toloka.TolokaClient,
    lang: str,
    pool_id: str,
) -> toloka.Assignment.Status:
    verdict: Optional[toloka.Assignment.Status] = None
    accuracy = evaluation.get_accuracy()
    for rule in block_rules:
        if not evaluation.assignment.id:
            continue  # model workers are not blocked
        restriction = rule.apply(
            value=accuracy,
            client=client,
            user_id=evaluation.assignment.user_id,
            pool_id=pool_id,
            block_start=evaluation.assignment.submitted,
            assignment_status=evaluation.assignment.status,
        )
        if restriction is not None:
            logger.debug(f'add restriction {restriction} for user {evaluation.assignment.user_id} by {rule.predicate}')
    for rule in set_verdict_rules:
        verdict = rule.apply(
            value=accuracy,
            client=client,
            assignment=evaluation.assignment,
            public_comment=get_assignment_rejection_comment(evaluation.incorrect_solution_indexes, lang),
        )
        if not evaluation.assignment.id:
            # model worker assignments are not present in Toloka storage, so we have no possibility to store their
            # status as for usual workers, so we set it on the fly
            evaluation.assignment.status = verdict
        if verdict is not None:
            logger.debug(f'set assignment {evaluation.assignment.id} to status {verdict.value} by {rule.predicate}')
            break
    assert verdict is not None
    return verdict


assignment_short_rejection_comment = base.LocalizedString(
    {
        'EN': 'Too few correct solutions',
        'RU': 'Мало правильных решений',
        'TR': 'Çok az doğru karar var',
        'KK': 'Дұрыс шешімдер аз',
    }
)

assignment_rejection_comment = base.LocalizedString(
    {
        'EN': 'Check the tasks with numbers: {incorrect_solutions_str}. Learn more about tasks acceptance and filing '
        'appeals in the project instructions.',
        'RU': 'Проверьте задания c номерами: {incorrect_solutions_str}. Подробнее о проверке заданий и подаче '
        'апелляций читайте в инструкции проекта.',
        'TR': 'Numaralardaki görevleri kontrol edin: {incorrect_solutions_str}. Görevlerin incelenmesi ve temyiz '
        'başvurusu hakkında daha fazla bilgi için proje talimatlarını okuyun.',
        'KK': 'Тапсырмаларды нөмірлермен тексеріңіз: {incorrect_solutions_str}. Тапсырмаларды тексеру және '
        'апелляцияларды беру туралы толығырақ жоба нұсқаулығынан оқыңыз.',
    }
)


def get_assignment_rejection_comment(incorrect_solutions_indexes: List[int], lang: str) -> str:
    if incorrect_solutions_indexes:
        incorrect_solutions_str = ', '.join(str(i + 1) for i in incorrect_solutions_indexes)
        return assignment_rejection_comment[lang].format(incorrect_solutions_str=incorrect_solutions_str)
    return assignment_short_rejection_comment[lang]


def evaluate_submitted_assignments_and_apply_rules(
    submitted_assignments: List[mapping.AssignmentSolutions],
    assignment_accuracy_evaluation_strategy: AssignmentAccuracyEvaluationStrategy,
    control_params: control.Control,
    client: toloka.TolokaClient,
    lang: str,
    pool_id: str,
) -> List[toloka.Assignment.Status]:
    assert all(assignment.status == toloka.Assignment.SUBMITTED for assignment, _ in submitted_assignments)

    assignment_evaluations = [
        assignment_accuracy_evaluation_strategy.evaluate_assignment(assignment) for assignment in submitted_assignments
    ]

    set_verdict_rules = control_params.filter_rules(
        predicate_type=control.AssignmentAccuracyPredicate, action_type=control.SetAssignmentStatus
    )

    block_rules = control_params.filter_rules(
        predicate_type=control.AssignmentAccuracyPredicate, action_type=control.BlockUser
    )

    return [
        apply_rules_to_assignment(set_verdict_rules, block_rules, evaluation, client, lang, pool_id)
        for evaluation in assignment_evaluations
    ]


def get_markup_task_id(check_task_id: mapping.TaskID, markup_task_mapping: mapping.TaskMapping) -> mapping.TaskID:
    """
    We receive markup task evaluations as map markup_solution_id -> evaluation.
    Sometimes we need to get markup_task_id from markup_solution_id.
    """
    return markup_task_mapping.task_id(check_task_id.objects[: len(markup_task_mapping.input_mapping)])


# Задание финализируется, если соблюдено хотя бы одно из условий:
# 1) У него есть решение приемлемого качества.
# 2) Число попыток исчерпано.
# 3) Мы проверяем не все решения со страниц, задание со страницы не проверено,
#    но точность страницы достаточно высокая.
def find_finalized_tasks(
    markup_assignments: List[mapping.AssignmentSolutions],
    solution_id_to_evaluation: Dict[mapping.TaskID, SolutionEvaluation],
    markup_task_mapping: mapping.TaskMapping,
    check_task_mapping: mapping.TaskMapping,
    check_sample: Optional[AssignmentCheckSample],
    max_object_markup_attempts: Optional[int],
) -> Set[mapping.TaskID]:
    assignment_accuracy_evaluation_strategy = CheckAssignmentAccuracyEvaluationStrategy(
        solution_id_to_evaluation, check_task_mapping
    )
    assignments_evaluations = [
        assignment_accuracy_evaluation_strategy.evaluate_assignment(assignment) for assignment in markup_assignments
    ]
    task_id_to_markup_attempts = get_tasks_attempts(markup_assignments, markup_task_mapping)

    finalized_task_ids = set()
    for solution_id, solution_evaluation in solution_id_to_evaluation.items():
        if solution_evaluation.ok:
            finalized_task_ids.add(get_markup_task_id(solution_id, markup_task_mapping))

    for task_id, markup_attempts in task_id_to_markup_attempts.items():
        if max_object_markup_attempts is not None and markup_attempts >= max_object_markup_attempts:
            finalized_task_ids.add(task_id)

    assignment_id_to_accuracy = {
        evaluation.assignment.id: evaluation.get_accuracy() for evaluation in assignments_evaluations
    }

    for assignment, solutions in markup_assignments:
        assignment_accuracy = assignment_id_to_accuracy.get(assignment.id)
        if (
            assignment_accuracy is not None
            and check_sample is not None
            and check_sample.assignment_accuracy_finalization_threshold is not None
            and assignment_accuracy >= check_sample.assignment_accuracy_finalization_threshold
        ):
            for input_objects, output_objects in solutions:
                solution_id = check_task_mapping.task_id(input_objects + output_objects)
                if solution_id not in solution_id_to_evaluation:
                    finalized_task_ids.add(markup_task_mapping.task_id(input_objects))

    return finalized_task_ids


def get_ok_tasks_ratio(
    solution_id_to_evaluation: Dict[mapping.TaskID, SolutionEvaluation],
    markup_task_mapping: mapping.TaskMapping,
) -> float:
    tasks, ok_tasks = set(), set()
    for solution_id, solution_evaluation in solution_id_to_evaluation.items():
        task_id = get_markup_task_id(solution_id, markup_task_mapping)
        tasks.add(task_id)
        if solution_evaluation.ok:
            ok_tasks.add(task_id)
    return len(ok_tasks) / len(tasks)


def calculate_bonuses_for_markup_assignment(
    rules: List[control.Rule],
    assignment_evaluation: AssignmentEvaluation,
) -> List[toloka.user_bonus.UserBonus]:
    bonuses = []
    for rule in rules:
        bonus = rule.apply(
            user_id=assignment_evaluation.assignment.user_id,
            assignment_id=assignment_evaluation.assignment.id,
            value=assignment_evaluation.get_accuracy(),
        )
        if bonus is not None:
            bonuses.append(bonus)

    return bonuses


def calculate_bonuses_for_pool_of_markup_assignments(
    markup_assignments: List[mapping.AssignmentSolutions],
    solution_id_to_evaluation: Dict[mapping.TaskID, SolutionEvaluation],
    check_task_mapping: mapping.TaskMapping,
    control_params: control.Control,
) -> List[toloka.user_bonus.UserBonus]:
    assert all(
        assignment.status in (toloka.Assignment.ACCEPTED, toloka.Assignment.REJECTED)
        for assignment, _ in markup_assignments
    )

    # todo add ability to filter rules by only action/only predicate type
    bonus_rules = control_params.filter_rules(
        predicate_type=control.AssignmentAccuracyPredicate, action_type=control.GiveBonusToUser
    )

    if not bonus_rules:
        return []

    assignment_accuracy_evaluation_strategy = CheckAssignmentAccuracyEvaluationStrategy(
        solution_id_to_evaluation, check_task_mapping
    )
    assignment_evaluations = [
        assignment_accuracy_evaluation_strategy.evaluate_assignment(assignment) for assignment in markup_assignments
    ]

    bonuses = []
    for evaluation in assignment_evaluations:
        bonuses += calculate_bonuses_for_markup_assignment(bonus_rules, evaluation)

    return bonuses


def calculate_worker_weights(
    assignments: List[mapping.AssignmentSolutions],
    assignment_accuracy_evaluation_strategy: Optional[AssignmentAccuracyEvaluationStrategy],
    smoothness_k: float = 0.5,
) -> Dict[str, float]:
    assignment_evaluations = [
        assignment_accuracy_evaluation_strategy.evaluate_assignment(assignment) for assignment in assignments
    ]
    correct_count = defaultdict(int)
    total_count = defaultdict(int)
    for assignment_evaluation in assignment_evaluations:
        worker_id = assignment_evaluation.assignment.user_id
        correct_count[worker_id] += assignment_evaluation.ok_checks
        total_count[worker_id] += assignment_evaluation.total_checks

    return {
        worker_id: (correct + smoothness_k) / (total_count[worker_id] + 2 * smoothness_k)
        for worker_id, correct in correct_count.items()
    }



import abc
from collections import defaultdict
from datetime import datetime, timedelta
import logging
import threading
from time import sleep
from typing import List, Dict, Optional, Tuple, Any

from dataclasses import dataclass, field

import pandas as pd
from IPython.display import display, clear_output, Image
import toloka.client as toloka

from .. import base, classification, classification_loop, client, evaluation, feedback_loop, mapping

logger = logging.getLogger(__name__)


def get_durations(assignment_solutions: List[mapping.AssignmentSolutions]) -> List[timedelta]:
    return [
        (assignment.submitted - assignment.created) / len(assignment.tasks)
        for assignment, _ in assignment_solutions
        if assignment.status != toloka.Assignment.REJECTED
    ]


@dataclass
class Metrics:
    durations: List[timedelta]
    overlaps: List[int]
    probas: List[float]
    confusion_matrix: Optional[pd.DataFrame] = None


def _collect_and_update_pool_metrics(
    toloka_client: toloka.TolokaClient,
    pool_id: str,
    history: Dict[str, List[int]],
) -> int:
    operation = toloka_client.get_analytics(
        [
            toloka.analytics_request.SubmitedAssignmentsCountPoolAnalytics(subject_id=pool_id),
            toloka.analytics_request.RejectedAssignmentsCountPoolAnalytics(subject_id=pool_id),
            toloka.analytics_request.ApprovedAssignmentsCountPoolAnalytics(subject_id=pool_id),
            toloka.analytics_request.SkippedAssignmentsCountPoolAnalytics(subject_id=pool_id),
        ]
    )
    operation = toloka_client.wait_operation(operation)

    for result in operation.details['value']:
        status = result['request']['name'].split('_')[0]
        history[status].append(result['result'])

    operation = toloka_client.get_analytics(
        [toloka.analytics_request.CompletionPercentagePoolAnalytics(subject_id=pool_id)]
    )
    operation = toloka_client.wait_operation(operation)
    completed = operation.details['value'][0]['result']['value']
    return completed


def _get_confusion_matrix(
    task_mapping: mapping.TaskMapping,
    assignment_solutions: List[mapping.AssignmentSolutions],
) -> pd.DataFrame:
    assert len(task_mapping.output_mapping) == 1
    label_class = task_mapping.output_mapping[0].obj_type
    assert issubclass(label_class, base.Label)
    label_set = label_class.possible_values()
    instance_to_value = {
        instance: value for instance, value in zip(label_class.possible_instances(), label_class.possible_values())
    }

    statistics = {real_label: {submitted_label: 0 for submitted_label in label_set} for real_label in label_set}
    for assignment, solutions in assignment_solutions:
        for task, solution in zip(assignment.tasks, solutions):
            if not task.known_solutions:
                continue
            assert len(task.known_solutions) == 1

            real_solution = task_mapping.from_solution(task.known_solutions[0])
            _, submitted_solution = solution
            statistics[instance_to_value[real_solution[0]]][instance_to_value[submitted_solution[0]]] += 1

    df = pd.DataFrame([statistics[label] for label in label_set], index=label_set)
    df.index.name = 'Real label'
    df.columns.name = 'Submitted label'
    return df


def _collect_and_update_classification_metrics(
    toloka_client: toloka.TolokaClient,
    task_mapping: mapping.TaskMapping,
    pool_id: str,
    assignment_evaluation_strategy: evaluation.AssignmentAccuracyEvaluationStrategy,
    aggregation_algorithm: classification.AggregationAlgorithm,
    history: Dict[str, List[int]],
    completed: List[int],
) -> Metrics:
    assignments_solutions = classification_loop.get_assignments_solutions(
        toloka_client,
        task_mapping,
        pool_id,
        [toloka.Assignment.ACCEPTED, toloka.Assignment.REJECTED, toloka.Assignment.SUBMITTED],
        with_control_tasks=True,
    )
    not_rejected_assignments_solutions = [
        (assignment, solution)
        for assignment, solution in assignments_solutions
        if assignment.status != toloka.Assignment.REJECTED
    ]
    durations = get_durations(assignments_solutions)

    task_id_to_label_confidence = classification_loop.calculate_label_probas(
        toloka_client,
        task_mapping,
        assignment_evaluation_strategy,
        aggregation_algorithm,
        assignments_solutions,
        pool_id,
    )
    task_id_to_current_overlap = evaluation.get_tasks_attempts(not_rejected_assignments_solutions, task_mapping)
    overlaps = [
        overlap for task_id, overlap in task_id_to_current_overlap.items() if task_id in task_id_to_label_confidence
    ]
    probas = [proba for _, proba in task_id_to_label_confidence.values()]
    completed.append(_collect_and_update_pool_metrics(toloka_client, pool_id, history))
    confusion_matrix = _get_confusion_matrix(task_mapping, not_rejected_assignments_solutions)
    return Metrics(durations, overlaps, probas, confusion_matrix)


def _collect_and_update_markup_metrics(
    toloka_client: toloka.TolokaClient,
    check_task_mapping: mapping.TaskMapping,
    markup_task_mapping: mapping.TaskMapping,
    check_pool_id: str,
    markup_pool_id: str,
    aggregation_algorithm: classification.AggregationAlgorithm,
    check_assignment_evaluation_strategy: evaluation.AssignmentAccuracyEvaluationStrategy,
    history: Dict[str, Dict[str, List[int]]],
    completed: Dict[str, List[int]],
) -> Tuple[Metrics, Metrics]:

    check_assignments = classification_loop.get_assignments_solutions(
        toloka_client,
        check_task_mapping,
        check_pool_id,
        [toloka.Assignment.ACCEPTED, toloka.Assignment.REJECTED, toloka.Assignment.SUBMITTED],
        with_control_tasks=True,
    )

    check_durations = get_durations(check_assignments)

    not_rejected_check_assignments_solutions = [
        assignment_solution
        for assignment_solution in check_assignments
        if assignment_solution[0].status != toloka.Assignment.REJECTED
    ]

    check_worker_weights = (
        evaluation.calculate_worker_weights(check_assignments, check_assignment_evaluation_strategy)
        if aggregation_algorithm == classification.AggregationAlgorithm.MAX_LIKELIHOOD
        else None
    )

    check_task_id_to_label_confidence = classification_loop.calculate_label_probas(
        toloka_client,
        check_task_mapping,
        check_assignment_evaluation_strategy,
        aggregation_algorithm,
        check_assignments,
        check_pool_id,
    )

    check_overlaps = [
        overlap
        for task_id, overlap in evaluation.get_tasks_attempts(
            not_rejected_check_assignments_solutions, check_task_mapping
        ).items()
        if task_id in check_task_id_to_label_confidence
    ]
    check_probas = [proba for _, proba in check_task_id_to_label_confidence.values()]

    markup_assignments = classification_loop.get_assignments_solutions(
        toloka_client,
        markup_task_mapping,
        markup_pool_id,
        [toloka.Assignment.ACCEPTED, toloka.Assignment.REJECTED, toloka.Assignment.SUBMITTED],
    )

    markup_durations = get_durations(markup_assignments)

    markup_solution_id_to_evaluation = evaluation.collect_evaluations_from_check_assignments(
        assignments=list(toloka_client.get_assignments(status=toloka.Assignment.ACCEPTED, pool_id=check_pool_id)),
        check_task_mapping=check_task_mapping,
        pool_input_objects=None,
        aggregation_algorithm=aggregation_algorithm,
        confidence_threshold=0.0,
        worker_weights=check_worker_weights,
    )  # todo здесь вроде неважно, какой confidence использовать
    markup_task_id_to_attempts_count = evaluation.get_tasks_attempts(markup_assignments, markup_task_mapping)
    markup_overlaps = list(markup_task_id_to_attempts_count.values())
    markup_probas = [
        solution_evaluation.confidence for solution_evaluation in markup_solution_id_to_evaluation.values()
    ]

    completed['check'].append(_collect_and_update_pool_metrics(toloka_client, check_pool_id, history['check']))
    completed['annotation'].append(
        _collect_and_update_pool_metrics(toloka_client, markup_pool_id, history['annotation'])
    )
    return Metrics(markup_durations, markup_overlaps, markup_probas), Metrics(
        check_durations, check_overlaps, check_probas
    )


def _plot_timeline(ax, times: List[datetime], history: Dict[str, List[int]], pool_name: str = ''):
    for status, history_values in history.items():
        ax.plot(times, history_values, label=status)
    ax.legend(loc=4, fontsize=16)
    ax.set_title(f'{pool_name} task assignments timeline'.strip().capitalize(), size=24)
    ax.tick_params(labelsize=16)


def _plot_completion_timeline(ax, times: List[datetime], completed: List[int], pool_name: str = ''):
    ax.plot(times, completed, label='%')
    ax.legend(loc=4, fontsize=16)
    ax.set_title(f'{pool_name} completion timeline'.strip().capitalize(), size=24)
    ax.tick_params(labelsize=16)


def _plot_time_distribution(ax, durations: List[timedelta], task_duration_hint: timedelta, pool_name: str = ''):
    ax.axvline(task_duration_hint.total_seconds(), c='red', label='hint', linestyle='--')
    ax.hist([duration.total_seconds() for duration in durations], alpha=0.3)
    ax.set_title(f'{pool_name} task duration distribution (seconds)'.strip().capitalize(), size=24)
    ax.legend(loc=4, fontsize=16)
    ax.tick_params(labelsize=16)


def _plot_confidence_distribution(ax, probas: List[float], pool_name: str = ''):
    ax.hist(probas, alpha=0.3)
    ax.set_title(f'{pool_name} confidence distribution'.strip().capitalize(), size=24)
    ax.tick_params(labelsize=16)


def _plot_attempts_distribution(ax, overlaps: List[int], pool_name: str = ''):
    ax.hist(overlaps, alpha=0.3)
    ax.set_title(f'{pool_name} overlap distribution'.strip().capitalize(), size=24)
    ax.tick_params(labelsize=16)


def _plot_confusion_matrix(ax, confusion_matrix: Optional[pd.DataFrame], pool_name: str = ''):
    if confusion_matrix is None:
        return
    try:
        from seaborn import heatmap

        heatmap(
            confusion_matrix,
            ax=ax,
            vmin=0,
            annot=True,
            fmt="d",
            linewidths=0.5,
            cmap='Blues',
            annot_kws={'size': 14},
        )
        ax.tick_params(labelsize=16)
        ax.set_title(f'{pool_name} control tasks confusion matrix'.strip().capitalize(), size=24)
    except ImportError:
        return


@dataclass
class MetricsPlotter:
    toloka_client: toloka.TolokaClient
    stop_event: threading.Event
    thread: threading.Thread = field(init=False)
    output: Any = field(init=False)

    def __post_init__(self):
        from ipywidgets import Output

        self.plots_image_file_name = 'metrics.png'
        self.redraw_period_seconds = 60

        self.output = Output()
        display(self.output)
        self.thread = threading.Thread(target=self.plot)
        self.thread.start()

    @abc.abstractmethod
    def plot(self):
        ...

    def post_plot(self) -> bool:
        import matplotlib.pyplot as plt

        plt.savefig(self.plots_image_file_name)
        plt.close()

        with self.output:
            clear_output(wait=True)
            display(Image(self.plots_image_file_name))

        logger.debug('metrics are updated')

        if self.stop_event.is_set():
            logger.debug('terminating metrics plotter')
            return True

        sleep(self.redraw_period_seconds)
        return False

    def join(self):
        self.thread.join()


@dataclass
class ClassificationMetricsPlotter(MetricsPlotter):
    task_mapping: mapping.TaskMapping
    pool_id: str
    task_duration_hint: timedelta
    params: classification_loop.Params
    assignment_evaluation_strategy: evaluation.AssignmentAccuracyEvaluationStrategy

    def plot(self):
        import matplotlib.pyplot as plt

        history = defaultdict(list)
        completed = []
        times = []
        aggregation_algorithm = self.params.aggregation_algorithm

        while True:
            metrics = _collect_and_update_classification_metrics(
                self.toloka_client,
                self.task_mapping,
                self.pool_id,
                self.assignment_evaluation_strategy,
                aggregation_algorithm,
                history,
                completed,
            )
            times.append(datetime.now())
            with plt.style.context({'axes.labelsize': 20}):
                fig, ax = plt.subplots(2, 3, figsize=(30, 15))
                _plot_timeline(ax[0][0], times, history)
                _plot_completion_timeline(ax[0][1], times, completed)
                _plot_time_distribution(ax[0][2], metrics.durations, self.task_duration_hint)
                _plot_confidence_distribution(ax[1][0], metrics.probas)
                _plot_attempts_distribution(ax[1][1], metrics.overlaps)
                _plot_confusion_matrix(ax[1][2], metrics.confusion_matrix)
            stopped = self.post_plot()
            if stopped:
                return


@dataclass
class AnnotationMetricsPlotter(MetricsPlotter):
    task_spec: client.AnnotationTaskSpec
    check_pool_id: str
    markup_pool_id: str
    check_task_duration_hint: timedelta
    markup_task_duration_hint: timedelta
    evaluation: feedback_loop.Evaluation

    def plot(self):
        import matplotlib.pyplot as plt

        history = defaultdict(lambda: defaultdict(list))
        completed = defaultdict(list)
        times = []
        assert isinstance(self.task_spec.function, base.AnnotationFunction)
        check_task_mapping, markup_task_mapping = self.task_spec.check.task_mapping, self.task_spec.task_mapping
        check_assignment_evaluation_strategy = evaluation.ControlTasksAssignmentAccuracyEvaluationStrategy(
            check_task_mapping
        )
        aggregation_algorithm = self.evaluation.aggregation_algorithm

        while True:
            markup_metrics, check_metrics = _collect_and_update_markup_metrics(
                self.toloka_client,
                check_task_mapping,
                markup_task_mapping,
                self.check_pool_id,
                self.markup_pool_id,
                aggregation_algorithm,
                check_assignment_evaluation_strategy,
                history,
                completed,
            )

            times.append(datetime.now())
            fig, ax = plt.subplots(3, 3, figsize=(30, 30))
            for i, (pool_name, metrics, hint) in enumerate(
                [
                    ('annotation', markup_metrics, self.markup_task_duration_hint),
                    ('check', check_metrics, self.check_task_duration_hint),
                ]
            ):
                _plot_timeline(ax[i][0], times, history[pool_name], pool_name=pool_name)
                _plot_completion_timeline(ax[i][1], times, completed[pool_name], pool_name=pool_name)
                _plot_time_distribution(ax[i][2], metrics.durations, hint, pool_name=pool_name)
                _plot_attempts_distribution(ax[2][i + 1], metrics.overlaps, pool_name=pool_name)
                if pool_name == 'annotation':
                    _plot_confidence_distribution(ax[2][0], metrics.probas, pool_name=pool_name)

            stopped = self.post_plot()
            if stopped:
                return



from dataclasses import dataclass
from enum import Enum
import math
import json
from typing import List, Optional, Type, Tuple, ClassVar, Dict, Union, Literal

import numpy as np
import pandas as pd

from .. import (
    base,
    classification,
    classification_loop,
    contrib,
    evaluation,
    feedback_loop,
    mapping,
    objects,
    pricing,
    project,
)

from .task import PreparedTaskSpec, AnnotationTaskSpec, PreviewTaskSpec, AnnotationPreviewTaskSpec
from ..base import TaskSpec
from ..worker import NoWorker
from ..utils import DecimalEncoder


def get_pool_price(
    input_objects_count: int,
    config: pricing.PoolPricingConfig,
    overlap: classification_loop.Overlap,
    display_formula: bool = False,
) -> str:
    formula = PoolPriceFormula(input_objects_count, config, overlap)
    if display_formula:
        from IPython.display import display, Math

        print('clear formula, which does not account edge cases like min commission and incomplete assignments')
        display(Math(formula.clear_formula()))  # noqa
        print('\nmore precise formula, which accounts more edge cases')
        display(Math(formula.precise_formula()))  # noqa
    return formula.total_price_precise_raw_str


def get_annotation_price(
    input_objects_count: int,
    markup_config: pricing.PoolPricingConfig,
    check_config: pricing.PoolPricingConfig,
    markup_overlap: classification_loop.DynamicOverlap,
    check_overlap: classification_loop.Overlap,
    assignment_check_sample: Optional[evaluation.AssignmentCheckSample],
    display_formula: bool = False,
) -> str:
    formula = AnnotationPriceFormula(
        input_objects_count=input_objects_count,
        markup_config=markup_config,
        check_config=check_config,
        markup_overlap=markup_overlap,
        check_overlap=check_overlap,
        assignment_check_sample=assignment_check_sample,
    )
    if display_formula:
        from IPython.display import display, Math

        print('some edge cases may be not fully accounted')
        display(*(Math(f) for f in formula.clear_formula()))
    return formula.total_price_simple_raw_str


def get_spec_cls(task_function: base.TaskFunction, for_preview: bool = False) -> Type[PreparedTaskSpec]:
    if type(task_function) in [base.ClassificationFunction, base.SbSFunction]:
        return PreviewTaskSpec if for_preview else PreparedTaskSpec
    elif isinstance(task_function, base.AnnotationFunction):
        return AnnotationPreviewTaskSpec if for_preview else AnnotationTaskSpec
    else:
        raise ValueError(f'Unknown task function: {task_function}')


def get_spec_cls_by_task_spec(task_spec: PreparedTaskSpec) -> Type[PreparedTaskSpec]:
    if isinstance(task_spec, AnnotationTaskSpec):
        return AnnotationPreviewTaskSpec
    return PreviewTaskSpec


# current formula is not 100% precise:
# - it does not know about custom toloka commission
# - it rounds total price by cents, but is seems that Toloka using ceil() for it
# also, we have two variants of formulas
# - clear one, which does not account case than assignment price is such that commission is greater than default
#   because of min commission, and also incomplete last assignment
# - precise but complex one, which accounts min commission
@dataclass
class PoolPriceFormula:
    input_objects_count: int
    config: pricing.PoolPricingConfig
    overlap: classification_loop.Overlap

    def __post_init__(self):
        min_overlap = self.overlap.min_overlap
        self.has_price_range = isinstance(self.overlap, classification_loop.DynamicOverlap)
        max_overlap = self.overlap.max_overlap if self.has_price_range else min_overlap

        self.task_price = self.config.assignment_price / self.config.real_tasks_count

        base_price_clear = self.input_objects_count * self.task_price * (1 + pricing.toloka_commission)

        self.min_total_price_clear = base_price_clear * min_overlap
        self.max_total_price_clear = base_price_clear * max_overlap

        self.assignments_count = math.ceil(self.input_objects_count / self.config.real_tasks_count)
        self.assignments_price = self.config.assignment_price + max(
            self.config.assignment_price * pricing.toloka_commission, pricing.min_toloka_commission
        )

        base_price_precise = self.assignments_count * self.assignments_price
        self.min_total_price_precise = base_price_precise * min_overlap
        self.max_total_price_precise = base_price_precise * max_overlap

        self.overlap_str = f'{self.overlap.min_overlap}'
        self.overlap_explanation_str = 'Overlap'
        self.total_price_simple_str = fr'{self.min_total_price_clear:.2f}\$'
        self.total_price_precise_str = fr'{self.min_total_price_precise:.2f}\$'
        self.total_price_precise_raw_str = f'{self.min_total_price_precise:.2f}$'

        if self.has_price_range:
            self.overlap_str = fr'[{self.overlap_str} \dots {self.overlap.max_overlap}]'
            self.overlap_explanation_str = r'[MinOverlap \dots MaxOverlap]'
            self.total_price_simple_str = fr'{self.total_price_simple_str} \dots {self.max_total_price_clear:.2f}\$'
            self.total_price_precise_str = fr'{self.total_price_precise_str} \dots {self.max_total_price_precise:.2f}\$'
            self.total_price_precise_raw_str = (
                f'{self.total_price_precise_raw_str} ... {self.max_total_price_precise:.2f}$'
            )

    def clear_formula(self) -> str:
        return (
            r'TotalPrice_{clear} = TaskCount * PricePerTask_\$ * '
            fr'{self.overlap_explanation_str} * (1 + TolokaCommission) = '
            fr'{self.input_objects_count} * {self.task_price:.4f}\$ * {self.overlap_str} * '
            fr'{1 + pricing.toloka_commission} = {self.total_price_simple_str}.'
        )

    def precise_formula(self) -> str:
        return (
            r'TotalPrice_{precise} = AssignmentCount * PricePerAssignment_\$ * '
            fr'{self.overlap_explanation_str} = \left \lceil \frac {{TaskCount}} {{TasksOnAssignment}} \right \rceil * '
            r'(PricePerAssignment_\$ + max(PricePerAssignment_\$ * TolokaCommission, MinTolokaCommission_\$) '
            fr'* {self.overlap_explanation_str} = \lceil {self.input_objects_count} / '
            fr'{self.config.real_tasks_count} \rceil * '
            fr'({self.config.assignment_price}\$ + max({self.config.assignment_price}\$ * '
            fr'{pricing.toloka_commission}, {pricing.min_toloka_commission}\$) * {self.overlap_str} = '
            fr'{self.assignments_count} * {self.assignments_price:.3f} * {self.overlap_str} = '
            fr'{self.total_price_precise_str}.'
        )


@dataclass
class AnnotationPriceFormula:
    input_objects_count: int
    markup_config: pricing.PoolPricingConfig
    check_config: pricing.PoolPricingConfig

    markup_overlap: classification_loop.DynamicOverlap
    check_overlap: classification_loop.Overlap

    assignment_check_sample: Optional[evaluation.AssignmentCheckSample]

    def __post_init__(self):
        min_markup_overlap = self.markup_overlap.min_overlap
        max_markup_overlap = self.markup_overlap.max_overlap

        min_check_overlap = self.check_overlap.min_overlap
        self.has_check_price_range = isinstance(self.check_overlap, classification_loop.DynamicOverlap)
        max_check_overlap = self.check_overlap.max_overlap if self.has_check_price_range else min_check_overlap

        self.markup_task_price = self.markup_config.assignment_price / self.markup_config.real_tasks_count
        self.check_task_price = self.check_config.assignment_price / self.check_config.real_tasks_count

        self.check_overlap_str = f'{self.check_overlap.min_overlap}'
        self.markup_overlap_str = fr'[{self.markup_overlap.min_overlap} \dots {self.markup_overlap.max_overlap}]'
        self.markup_overlap_explanation_str = r'[MinMarkupOverlap \dots MaxMarkupOverlap]'
        self.check_overlap_explanation_str = 'CheckOverlap'

        self.check_tasks_count_str = (
            fr'[{self.input_objects_count * min_markup_overlap} \dots {self.input_objects_count * max_markup_overlap}]'
        )
        self.check_tasks_count_explanation_str = (
            f'MarkupTaskCount * {self.markup_overlap_explanation_str} = '
            f'{self.input_objects_count} * {self.markup_overlap_str}'
        )

        sample_frac = 1.0
        if self.assignment_check_sample is not None:
            assert self.assignment_check_sample.max_tasks_to_check is not None
            sample_frac = self.assignment_check_sample.max_tasks_to_check / self.markup_config.real_tasks_count

            self.check_tasks_count_str = (
                fr'[{math.ceil(self.input_objects_count * min_markup_overlap * sample_frac)} \dots '
                f'{math.ceil(self.input_objects_count * max_markup_overlap * sample_frac)}]'
            )
            self.check_tasks_count_explanation_str = (
                r'MarkupTaskCount * \frac {CheckSampleTaskCount} {MarkupTasksPerAssignment} * '
                fr'{self.markup_overlap_explanation_str} = {self.input_objects_count} * \frac '
                f'{{{self.assignment_check_sample.max_tasks_to_check}}} {{{self.markup_config.real_tasks_count}}} * '
                f'{self.markup_overlap_str}'
            )

        self.check_comission = pricing.toloka_commission

        if self.check_config.assignment_price * pricing.toloka_commission < pricing.min_toloka_commission:
            self.check_comission = pricing.min_toloka_commission / self.check_config.assignment_price

        self.markup_comission = pricing.toloka_commission

        if self.markup_config.assignment_price * pricing.toloka_commission < pricing.min_toloka_commission:
            self.markup_comission = pricing.min_toloka_commission / self.markup_config.assignment_price

        self.check_comission_explanation_str = (
            r'max \left( TolokaCommission, \frac {MinTolokaCommission} {PricePerCheckAssignment} \right) = '
            fr'max \left( {pricing.toloka_commission}, \frac {{{pricing.min_toloka_commission}}}'
            fr'{{{self.check_config.assignment_price}}} \right)'
        )
        self.markup_comission_explanation_str = (
            r'max \left( TolokaCommission, \frac {MinTolokaCommission} {PricePerMarkupAssignment} \right) = '
            fr'max \left( {pricing.toloka_commission}, \frac {{{pricing.min_toloka_commission}}}'
            fr'{{{self.markup_config.assignment_price}}} \right)'
        )

        base_markup_price_clear = self.input_objects_count * self.markup_task_price * (1 + self.markup_comission)
        # todo: sample происходит из целого assignment'а, могут быть краевые эффекты - не абсолютно точная формула
        base_check_price_clear = (
            math.ceil(self.input_objects_count * sample_frac) * self.check_task_price * (1 + self.check_comission)
        )
        self.min_markup_price = base_markup_price_clear * min_markup_overlap
        self.max_markup_price = base_markup_price_clear * max_markup_overlap
        self.min_check_price = base_check_price_clear * min_check_overlap * min_markup_overlap
        self.max_check_price = base_check_price_clear * max_check_overlap * max_markup_overlap
        self.min_total_price_clear = self.min_markup_price + self.min_check_price
        self.max_total_price_clear = self.max_markup_price + self.max_check_price

        self.check_price_str = fr'{self.min_check_price:.2f}\$ \dots {self.max_check_price:.2f}\$'
        self.markup_price_str = fr'{self.min_markup_price:.2f}\$ \dots {self.max_markup_price:.2f}\$'
        self.total_price_simple_str = fr'{self.min_total_price_clear:.2f}\$ \dots {self.max_total_price_clear:.2f}\$'
        self.total_price_simple_raw_str = f'{self.min_total_price_clear:.2f}$ ... {self.max_total_price_clear:.2f}$'
        if self.has_check_price_range:
            self.check_overlap_str = fr'[{self.check_overlap_str} \dots {self.check_overlap.max_overlap}]'
            self.check_overlap_explanation_str = r'[MinCheckOverlap \dots MaxCheckOverlap]'

    def clear_formula(self) -> Tuple[str, ...]:
        return (
            f'MarkupTolokaComission = {self.markup_comission_explanation_str} = {self.markup_comission};',
            fr'MarkupPrice = MarkupTaskCount * PricePerMarkupTask_\$ * '
            fr'{self.markup_overlap_explanation_str} * (1 + MarkupTolokaComission) = '
            fr'{self.input_objects_count} * {self.markup_task_price:.4f}\$ * {self.markup_overlap_str} * '
            fr'{1 + self.markup_comission} = {self.markup_price_str};',
            f'CheckTaskCount = {self.check_tasks_count_explanation_str} = {self.check_tasks_count_str};',
            f'CheckTolokaComission = {self.check_comission_explanation_str} = {self.check_comission};',
            r'CheckPrice = CheckTaskCount * PricePerCheckTask_\$ * '
            fr'{self.check_overlap_explanation_str} * (1 + CheckTolokaComission) = '
            fr'{self.check_tasks_count_str} * {self.check_task_price:.4f}\$ * {self.check_overlap_str} * '
            fr'{1 + self.check_comission} = {self.check_price_str};',
            r'TotalPrice = MarkupPrice + CheckPrice = '
            fr'{self.markup_price_str} + {self.check_price_str} = {self.total_price_simple_str}.',
        )


# template builder saves config and input fields in encoded way to URL, but it doesn't save output fields and state of
# UI, so we can create preview just for task (without solution), and there will be unnecessary config block
class TaskPreview:
    def __init__(
        self,
        input_objects: mapping.Objects,
        *,
        task_spec: Optional[PreparedTaskSpec] = None,
        task_function: Optional[base.TaskFunction] = None,
        lang: Optional[str] = None,
        output_objects: Optional[mapping.Objects] = None,
    ):
        assert (
            task_spec is not None or task_function is not None and lang is not None
        ), 'Provide either task_spec or both task_function and language'
        for_preview = output_objects is not None
        if task_spec is None:
            task_spec = get_spec_cls(task_function, for_preview)(
                task_spec=TaskSpec(
                    id=None,
                    function=task_function,
                    name=base.EMPTY_STRING,
                    description=base.EMPTY_STRING,
                    instruction=None,
                ),
                lang=lang,
            )
        elif for_preview:
            task_spec = get_spec_cls_by_task_spec(task_spec)(
                task_spec=task_spec.task_spec,
                lang=task_spec.lang,
                scenario=task_spec.scenario,
            )
        self.url = self.get_url(input_objects, task_spec, output_objects)

    def get_link(self) -> str:
        return f'<a href="{self.url}" target="_blank" rel="noopener noreferrer">task preview</a>'

    def display_link(self):
        from IPython.display import HTML

        return HTML(self.get_link())

    def display(self):
        from IPython.display import HTML

        return HTML(
            f'<iframe src="{self.url}" onload="javascript:(function(o)'
            '{o.style.height=o.contentWindow.document.body.scrollHeight+"px";}(this));" '
            'style="height:800px;width:100%;border:none;overflow:hidden;"/>'
        )

    @classmethod
    def get_url(
        cls,
        input_objects: mapping.Objects,
        task_spec: PreparedTaskSpec,
        output_objects: Optional[mapping.Objects] = None,
    ) -> str:
        objects_count = len(input_objects)
        if output_objects is not None:
            # for results preview we display output types as inputs, due to tb limitations
            objects_count += len(output_objects)

        if isinstance(task_spec, AnnotationTaskSpec):
            if objects_count == len(task_spec.check.task_mapping.input_mapping):
                task_spec = task_spec.check
            else:
                assert objects_count == len(
                    task_spec.task_mapping.input_mapping
                ), 'incorrect number of input objects supplied'
        config, task_mapping = task_spec.dumped_view, task_spec.task_mapping
        preview_objects = input_objects
        if output_objects is not None:
            preview_objects = np.concatenate([input_objects, output_objects])
        input_values: dict = task_mapping.to_task(preview_objects).input_values

        dumped = json.dumps(
            {'config': config, 'input': json.dumps(input_values, ensure_ascii=False, cls=DecimalEncoder)},
            ensure_ascii=False,
        )
        compressed = contrib.LZString.compressToEncodedURIComponent(dumped)
        return f'https://tb.yandex.net/editor?config={compressed}'


class Results:
    row_input_objects: np.ndarray
    row_output_objects: np.ndarray

    task_spec: PreparedTaskSpec

    WORKER_FIELD: ClassVar[str] = 'worker'
    PREVIEW_FIELD: ClassVar[str] = 'preview'

    def _add_task_previews(self, df: pd.DataFrame) -> pd.DataFrame:
        urls = [
            TaskPreview(task_input_objects, output_objects=task_output_objects, task_spec=self.task_spec).get_link()
            for task_input_objects, task_output_objects in zip(
                self.row_input_objects[df.index], self.row_output_objects[df.index]
            )
        ]
        df_copy = df.copy()
        df_copy[self.PREVIEW_FIELD] = urls
        return df_copy

    def html_with_task_previews(self, df: pd.DataFrame):
        from IPython.display import HTML

        return HTML(self._add_task_previews(df).to_html(escape=False))

    @staticmethod
    def deduplicate(df: pd.DataFrame) -> pd.DataFrame:
        return df.loc[df.astype(str).drop_duplicates().index]


class LabelingResults(Results):
    df: pd.DataFrame
    has_worker_weights: bool
    RESULT_FIELD: ClassVar[str] = 'result'
    CONFIDENCE_FIELD: ClassVar[str] = 'confidence'


# currently we only have 1 use case for combined classes, and producing M * N proba items is excessive,
# because all possible answers for each question are pooled from same 5 options
# thus, currently we revert generated answers to their original form inside Results
def get_class_fields(cls_type: Type[base.Class]) -> List[Tuple[str, str]]:
    if issubclass(cls_type, objects.CombinedAnswer):
        cls_values = list(set(cls.get_original_answer().value for cls in cls_type.possible_instances()))
    else:
        cls_values = [cls.value for cls in cls_type.possible_instances()]

    return [(cls_value, f'proba_{cls_value}') for cls_value in cls_values]


def get_labels_probas(
    labels_probas: Optional[classification.TaskLabelsProbas], cls_type: Type[base.Class]
) -> Optional[classification.TaskLabelsProbas]:
    if labels_probas is None:
        return None
    if not issubclass(cls_type, objects.CombinedAnswer):
        return {key.value: value for key, value in labels_probas.items()}

    return {key.get_original_answer().value: value for key, value in labels_probas.items()}


def get_label(raw_label: Optional[base.Label]) -> Union[base.Label, Literal[np.nan]]:
    if raw_label is None:
        return np.nan
    if not isinstance(raw_label, objects.CombinedAnswer):
        return raw_label.value

    return raw_label.get_original_answer().value


class ClassificationResults(LabelingResults):
    WORKER_WEIGHT_FIELD: ClassVar[str] = 'worker_weight'
    LABEL_FIELD: ClassVar[str] = 'label'
    OVERLAP_FIELD: ClassVar[str] = 'overlap'

    def __init__(
        self,
        input_objects: List[mapping.Objects],
        results: List[Tuple[classification.TaskLabelsProbas, List[classification.WorkerLabel]]],
        task_spec: PreparedTaskSpec,
        worker_weights: Optional[Dict[str, float]] = None,
    ):
        self.task_spec = task_spec

        # we will index these input objects by dataframe row indexes
        row_input_objects, row_output_objects = [], []

        input_fields = []
        for obj_mapping in task_spec.task_mapping.input_mapping:
            for _, task_field in obj_mapping.obj_task_fields:
                input_fields.append(task_field)

        self.input_fields = input_fields

        cls_type: Type[base.Class] = task_spec.task_mapping.output_mapping[0].obj_meta.type
        cls_fields = get_class_fields(cls_type)

        self.proba_fields = [cls_field for _, cls_field in cls_fields]
        self.has_worker_weights = worker_weights is not None
        rows = []
        for task_input_objects, task_result in zip(input_objects, results):
            task_row = task_spec.task_mapping.toloka_values(task_input_objects)
            del task_row[mapping.TASK_ID_FIELD]

            labels_probas, raw_labels = task_result
            labels_probas = get_labels_probas(labels_probas, cls_type)
            overlap = len(raw_labels)

            most_probable_result = classification.get_most_probable_label(labels_probas)
            if most_probable_result is not None:
                task_result, proba = most_probable_result
            else:
                assert labels_probas is None
                assert not raw_labels

                # todo: maybe confidence should also be None
                task_result, proba, labels_probas = None, 0.0, {}
                raw_labels = [(None, NoWorker())]
                if self.has_worker_weights:
                    worker_weights[None] = 0.0

            task_row[self.RESULT_FIELD] = np.nan if task_result is None else task_result
            task_row[self.CONFIDENCE_FIELD] = proba
            task_row[self.OVERLAP_FIELD] = overlap
            for cls_value, cls_field in cls_fields:
                proba = labels_probas.get(cls_value, 0.0)
                task_row[cls_field] = proba

            for raw_label, worker in raw_labels:
                worker_row = dict(task_row)
                worker_row[self.LABEL_FIELD] = get_label(raw_label)
                worker_row[self.WORKER_FIELD] = worker.id
                if self.has_worker_weights:
                    worker_row[self.WORKER_WEIGHT_FIELD] = worker_weights[worker.id]
                rows.append(worker_row)
                row_input_objects.append(task_input_objects)
                row_output_objects.append((raw_label,))

        self.df = pd.DataFrame.from_dict(rows)
        self.row_input_objects = np.array(row_input_objects)
        self.row_output_objects = np.array(row_output_objects)

    def predict(self) -> pd.DataFrame:
        return self.deduplicate(
            self.df[self.input_fields + [self.RESULT_FIELD, self.CONFIDENCE_FIELD, self.OVERLAP_FIELD]]
        )

    def predict_proba(self) -> pd.DataFrame:
        return self.deduplicate(self.df[self.input_fields + self.proba_fields + [self.OVERLAP_FIELD]])

    def worker_labels(self) -> pd.DataFrame:
        worker_labels = self.df[
            self.input_fields
            + [self.LABEL_FIELD, self.WORKER_FIELD]
            + ([self.WORKER_WEIGHT_FIELD] if self.has_worker_weights else [])
        ]
        return worker_labels[worker_labels[self.WORKER_FIELD].notnull()]


class AnnotationResults(LabelingResults):
    EVAL_FIELD: ClassVar[str] = base.EVALUATION_TASK_FIELD
    ANNOTATION_WORKER_FIELD: ClassVar[str] = 'annotator'
    CHECK_WORKER_FIELD: ClassVar[str] = 'evaluator'
    WORKER_WEIGHT_FIELD: ClassVar[str] = 'evaluator_weight'
    CHECK_OVERLAP_FIELD: ClassVar[str] = 'evaluation_overlap'
    ANNOTATION_OVERLAP_FIELD: ClassVar[str] = 'annotation_overlap'

    def __init__(
        self,
        input_objects: List[mapping.Objects],
        results: List[List[feedback_loop.Solution]],
        task_spec: AnnotationTaskSpec,
        worker_weights: Optional[Dict[str, float]] = None,
    ):
        self.task_spec = task_spec

        row_input_objects, row_output_objects = [], []

        input_fields = []
        for obj_mapping in task_spec.task_mapping.input_mapping:
            for _, task_field in obj_mapping.obj_task_fields:
                input_fields.append(task_field)

        self.input_fields = input_fields

        output_fields = []
        for obj_mapping in task_spec.task_mapping.output_mapping:
            for _, task_field in obj_mapping.obj_task_fields:
                output_fields.append(task_field)

        self.output_fields = output_fields

        self.has_worker_weights = worker_weights is not None

        rows = []
        for task_input_objects, solutions in zip(input_objects, results):
            inputs = task_spec.task_mapping.toloka_values(task_input_objects)
            del inputs[mapping.TASK_ID_FIELD]

            for i, solution in enumerate(solutions):
                outputs = task_spec.task_mapping.toloka_values(solution.solution, output=True)

                task_row = {**inputs, **outputs}
                task_row[self.ANNOTATION_WORKER_FIELD] = solution.worker.id

                task_row[self.RESULT_FIELD] = i == 0
                task_row[self.ANNOTATION_OVERLAP_FIELD] = len(solutions)
                if solution.evaluation is None:
                    task_row[self.CHECK_OVERLAP_FIELD] = 0
                    rows.append(task_row)
                    row_input_objects.append(task_input_objects)
                    row_output_objects.append(solution.solution)
                else:
                    assert solution.evaluation.worker_labels
                    task_row[self.CONFIDENCE_FIELD] = solution.evaluation.confidence
                    task_row[self.CHECK_OVERLAP_FIELD] = len(solution.evaluation.worker_labels)
                    for choice, worker in solution.evaluation.worker_labels:
                        worker_row = dict(task_row)
                        worker_row[self.EVAL_FIELD] = choice.ok
                        worker_row[self.CHECK_WORKER_FIELD] = worker.id
                        if self.has_worker_weights:
                            worker_row[self.WORKER_WEIGHT_FIELD] = worker_weights[worker.id]
                        rows.append(worker_row)
                        row_input_objects.append(task_input_objects)
                        row_output_objects.append(solution.solution)

        self.df = pd.DataFrame.from_dict(rows)
        self.row_input_objects = np.array(row_input_objects)
        self.row_output_objects = np.array(row_output_objects)

    def predict(self) -> pd.DataFrame:
        df = self.deduplicate(self.df[self.input_fields + self.output_fields + [self.RESULT_FIELD]])
        return df[df[self.RESULT_FIELD]].drop([self.RESULT_FIELD], axis=1)

    def predict_proba(self) -> pd.DataFrame:
        return self.deduplicate(self.df[self.input_fields + self.output_fields + [self.CONFIDENCE_FIELD]])

    def worker_labels(self) -> pd.DataFrame:
        return self.df.drop([self.RESULT_FIELD], axis=1)


class ExpertLabelingApplication(Enum):
    TRAINING = 'training'
    CONTROL_TASKS = 'control_tasks'
    ANNOTATION_CHECK_TRAINING = 'check_training'


class ExpertLabelingResults(Results):
    df: pd.DataFrame

    OK_FIELD: ClassVar[str] = '_ok'

    # todo implement this in ClassificationResults
    DURATION_FIELD: ClassVar[str] = 'duration'

    input_length: int
    output_lenght: int
    ok_index: int
    evaluation_index: Optional[int] = None
    comment_index: int
    output_names: List[str]

    def __init__(
        self,
        results: List[Tuple[Union[mapping.TaskSingleSolution, mapping.Objects], mapping.TaskMultipleSolutions]],
        task_spec: PreparedTaskSpec,
        worker_id_to_name: Optional[Dict[str, str]] = None,
    ):
        self.task_spec = task_spec
        if worker_id_to_name is None:
            worker_id_to_name = {}
        # we will index these input objects by dataframe row indexes
        row_input_objects, row_output_object = [], []

        rows = []
        for _, (input_objects, solution) in results:
            if not solution:
                # expert labeling was interrupted, and this task doesn't have a solution
                continue
            [(output_objects, assignment)] = solution
            task_row = {
                **task_spec.task_mapping.toloka_values(input_objects, output=False),
                **task_spec.task_mapping.toloka_values(output_objects, output=True),
            }
            del task_row[mapping.TASK_ID_FIELD]
            task_row[self.WORKER_FIELD] = worker_id_to_name.get(assignment.user_id, assignment.user_id)
            task_row[self.DURATION_FIELD] = (assignment.submitted - assignment.created) / len(assignment.tasks)
            rows.append(task_row)
            row_input_objects.append(input_objects)
            row_output_object.append(output_objects)
        self.rows = rows
        self.df = pd.DataFrame.from_dict(rows)
        self.row_input_objects = np.array(row_input_objects)
        self.row_output_objects = np.array(row_output_object)
        function = task_spec.function

        default_spec = get_spec_cls(function)(task_spec=self.task_spec.task_spec, lang=self.task_spec.lang)

        self.input_length = len(default_spec.task_mapping.input_mapping)
        self.output_length = len(default_spec.task_mapping.output_mapping)
        ok_meta, comment_meta = project.Builder.EXPERT_LABELING_METAS

        self.output_names = [
            task_mapping.obj_task_fields[0][1] for task_mapping in self.task_spec.task_mapping.output_mapping
        ]
        self.ok_index = self.output_names.index(ok_meta.name)
        self.comment_index = self.output_names.index(comment_meta.name)
        if isinstance(function, base.AnnotationFunction):
            evaluation_name = function.get_evaluation_named_meta().name
            if evaluation_name in self.output_names:
                self.evaluation_index = self.output_names.index(evaluation_name)

    def get_results(self) -> pd.DataFrame:
        return self.df

    def get_accuracy(self):
        return self.df[self.OK_FIELD].mean()

    def get_solved_task_and_comment(
        self,
        task: mapping.Objects,
        solution: mapping.Objects,
        application: ExpertLabelingApplication,
    ) -> Tuple[mapping.TaskSingleSolution, objects.Text]:
        objs = task + solution
        if self.evaluation_index is None or application == ExpertLabelingApplication.TRAINING:
            inputs, outputs, comment = (
                objs[: self.input_length],
                objs[self.input_length : self.input_length + self.output_length],
                solution[self.comment_index],
            )
        else:
            inputs, outputs, comment = (
                objs[: self.input_length + self.output_length],
                (solution[self.evaluation_index],),
                solution[self.comment_index],
            )
        return (inputs, outputs), comment

    def suitable(self, solution: mapping.Objects, application: ExpertLabelingApplication) -> bool:
        if not solution[self.ok_index].ok:
            return False
        if application == ExpertLabelingApplication.CONTROL_TASKS:
            return True
        if solution[self.comment_index] is None or not solution[self.comment_index].text:
            return False
        if application == ExpertLabelingApplication.ANNOTATION_CHECK_TRAINING:
            return True
        if self.evaluation_index is not None:
            return solution[self.evaluation_index].ok
        return True

    def get_correct_objects(
        self,
        application: ExpertLabelingApplication = ExpertLabelingApplication.CONTROL_TASKS,
    ) -> Tuple[List[mapping.TaskSingleSolution], List[objects.Text]]:
        tasks_and_solutions = [
            (self.task_spec.task_mapping.from_task_values(row), self.task_spec.task_mapping.from_solution_values(row))
            for row in self.rows
        ]
        generated_objects = [
            self.get_solved_task_and_comment(task, solution, application)
            for task, solution in tasks_and_solutions
            if self.suitable(solution, application)
        ]
        solutions = [solution for solution, _ in generated_objects]
        texts = [text for _, text in generated_objects]
        return solutions, texts



from dataclasses import dataclass, field
import logging
from typing import List, Dict, Union, Optional, Tuple

import toloka.client as toloka

from .. import base, classification, control, duration, evaluation, mapping, mos, pool as pool_config, utils, worker

logger = logging.getLogger(__name__)


@dataclass
class Overlap:
    min_overlap: int


@dataclass
class StaticOverlap(Overlap):
    overlap: int
    min_overlap: int = field(init=False, repr=False)

    def __post_init__(self):
        self.min_overlap = self.overlap

    def __hash__(self) -> int:
        return self.overlap

    @property
    def max_overlap(self) -> int:
        return self.overlap


@dataclass
class DynamicOverlap(Overlap):
    min_overlap: int
    max_overlap: int
    confidence: Union[float, classification.TaskLabelsProbas]

    def min_confidence(self, label: base.Label) -> float:
        if isinstance(self.confidence, float):
            return self.confidence
        return self.confidence[label]

    def __hash__(self) -> int:
        return hash(repr(self))


@dataclass
class Params:
    overlap: Overlap
    control: control.Control
    task_duration_function: duration.TaskDurationFunction
    aggregation_algorithm: Optional[classification.AggregationAlgorithm] = None


class ClassificationLoop:
    client: toloka.TolokaClient
    task_mapping: mapping.TaskMapping
    params: Params
    lang: str
    interrupter: Optional[utils.LaunchInterrupter]
    assignment_evaluation_strategy: evaluation.AssignmentAccuracyEvaluationStrategy
    model_ws: Optional[worker.ModelWorkspace]

    def __init__(
        self,
        client: toloka.TolokaClient,
        task_mapping: mapping.TaskMapping,
        params: Params,
        lang: str,
        interrupter: Optional[utils.LaunchInterrupter] = None,
        with_control_tasks: bool = True,
        model: Optional[worker.Model] = None,
    ):
        self.client = client
        self.task_mapping = task_mapping
        self.params = params
        self.lang = lang
        self.interrupter = interrupter
        self.assignment_evaluation_strategy = evaluation.ControlTasksAssignmentAccuracyEvaluationStrategy(
            self.task_mapping,
            can_have_zero_checks=not with_control_tasks,
        )
        self.model_ws = None
        if model:
            self.model_ws = worker.ModelWorkspace(model=model, task_mapping=self.task_mapping)

    def create_pool(
        self,
        control_objects: List[mapping.TaskSingleSolution],
        pool_cfg: pool_config.ClassificationConfig,
        check_project: bool = True,  # todo tmp for old pipelines
    ) -> toloka.Pool:
        if check_project:
            assert mapping.check_project_is_suitable(self.client.get_project(pool_cfg.project_id), self.task_mapping)
        assert pool_cfg.overlap == self.params.overlap.min_overlap
        pool = self.client.create_pool(pool_config.create_pool_params(pool_cfg))
        control_tasks = []
        for objects in control_objects:
            task = self.task_mapping.to_control_task(objects)
            task.pool_id = pool.id
            control_tasks.append(task)
        logger.debug(f'creating {len(control_tasks)} control tasks')
        if control_tasks:
            # call with empty task list will fail
            self.client.create_tasks(control_tasks, allow_defaults=True, async_mode=True, skip_invalid_items=False)
        return pool

    def add_input_objects(
        self, pool_id: str, input_objects: Union[List[mapping.Objects], List[Tuple[mapping.Objects, worker.Worker]]]
    ):
        tasks = []
        for objects_data in input_objects:
            objects, user_id = objects_data, None
            if len(objects_data) == 2 and isinstance(objects_data[1], worker.Worker):
                objects = objects_data[0]
                if isinstance(objects_data[1], worker.Human):
                    user_id = [objects_data[1].user_id]
            task = self.task_mapping.to_task(objects)
            task.unavailable_for = user_id
            task.pool_id = pool_id
            tasks.append(task)
        logger.debug(f'creating {len(tasks)} tasks')
        self.client.create_tasks(tasks, allow_defaults=True, async_mode=True, skip_invalid_items=False)
        if not self.model_ws:
            # in case of model worker, pool is only needed to store tasks
            self.client.open_pool(pool_id)

    def prior_filtration_is_enabled(self) -> bool:
        return True

    def loop(self, pool_id: str):
        if self.model_ws:
            return

        iteration = 1
        while True:
            logger.debug(f'classification loop iteration #{iteration} is started')
            utils.wait_pool_for_close(self.client, pool_id, self.interrupter)

            # TODO: collect stats about how workers answers control tasks and ignore bad control tasks by percentile

            submitted_assignments = self.get_assignments_solutions(
                pool_id, toloka.Assignment.SUBMITTED, with_control_tasks=True
            )

            logger.debug(f'{len(submitted_assignments)} submitted assignments are received')

            if self.prior_filtration_is_enabled():

                filtered_assignments, _ = evaluation.prior_filter_assignments(
                    self.client,
                    submitted_assignments,
                    self.params.control,
                    self.lang,
                    task_duration_function=self.params.task_duration_function,
                )
                logger.debug(f'{len(filtered_assignments)} assignments left after prior filter')

            else:
                filtered_assignments = submitted_assignments
                logger.debug('prior filtration is disabled')

            if isinstance(self.assignment_evaluation_strategy, evaluation.CustomEvaluationStrategy):
                self.assignment_evaluation_strategy.update(submitted_assignments)

            evaluation.evaluate_submitted_assignments_and_apply_rules(
                filtered_assignments,
                self.assignment_evaluation_strategy,
                self.params.control,
                self.client,
                self.lang,
                pool_id,
            )

            tasks_to_rework = rework_not_finalized_tasks(
                self.client,
                self.get_assignments_solutions(pool_id, [toloka.Assignment.ACCEPTED, toloka.Assignment.REJECTED]),
                self.task_mapping,
                # possible cases:
                # I. static overlap - increase until needed min_overlap is reached
                # II. dynamic overlap -
                #   1. increase until needed min_overlap is reached
                #   2. increase after new accepted solution, because not enough confidence accumulated
                #   3. NOT increase after new accepted solution, because enough confidence accumulated already
                #   4. increase after new rejected solution, because not enough confidence accumulated
                #   5. NOT increase after new rejected solution, because enough confidence accumulated due to
                #        confidence recalculation because worker weights are also recalculated
                self.get_task_id_to_overlap_increase(pool_id),
                pool_id,
            )
            if tasks_to_rework == 0:
                return

            iteration += 1

    def calculate_label_probas(
        self,
        assignment_solutions: List[mapping.AssignmentSolutions],
        pool_id: str,
    ) -> Dict[mapping.TaskID, classification.LabelProba]:
        return calculate_label_probas(
            self.client,
            self.task_mapping,
            self.assignment_evaluation_strategy,
            self.params.aggregation_algorithm,
            assignment_solutions,
            pool_id,
        )

    def get_task_id_to_overlap_increase(self, pool_id: str) -> Dict[mapping.TaskID, int]:
        # todo: we maybe should use _all_ assignments in dynamic overlap proba calculation, even rejected ones
        assignments_solutions = get_assignments_solutions(
            self.client,
            self.task_mapping,
            pool_id,
            [toloka.Assignment.ACCEPTED, toloka.Assignment.REJECTED],
            with_control_tasks=True,
        )
        accepted_assignments_solutions = [
            assignment_solution
            for assignment_solution in assignments_solutions
            if assignment_solution[0].status == toloka.Assignment.ACCEPTED
        ]
        task_id_to_current_overlap = evaluation.get_tasks_attempts(accepted_assignments_solutions, self.task_mapping)
        use_dynamic = isinstance(self.params.overlap, DynamicOverlap)
        task_id_to_label_confidence = self.calculate_label_probas(assignments_solutions, pool_id) if use_dynamic else {}
        task_ids = get_task_id_map(self.client, pool_id, self.task_mapping).keys()
        result = {}
        for task_id in task_ids:
            overlap = task_id_to_current_overlap[task_id]
            residual_overlap = self.params.overlap.min_overlap - overlap
            if residual_overlap > 0:
                result[task_id] = residual_overlap
            elif use_dynamic:
                assert isinstance(self.params.overlap, DynamicOverlap)
                label, conf = task_id_to_label_confidence[task_id]
                if overlap < self.params.overlap.max_overlap and conf < self.params.overlap.min_confidence(label):
                    result[task_id] = 1
        return result

    def get_assignments_solutions(
        self,
        pool_id: str,
        status: List[toloka.Assignment.Status],
        with_control_tasks: bool = False,
    ) -> List[mapping.AssignmentSolutions]:
        return get_assignments_solutions(self.client, self.task_mapping, pool_id, status, with_control_tasks)

    def get_results(
        self,
        pool_id: str,
        pool_input_objects: Optional[List[mapping.Objects]] = None,
    ) -> Tuple[
        List[Tuple[Optional[classification.TaskLabelsProbas], List[classification.WorkerLabel]]],
        Optional[Dict[str, float]],
    ]:
        pool_input_objects, accepted_assignments, worker_weights = self.get_assignments_and_worker_weights(
            pool_id, pool_input_objects
        )

        return (
            classification.collect_labels_probas_from_assignments(
                assignments=accepted_assignments,
                task_mapping=self.task_mapping,
                pool_input_objects=pool_input_objects,
                aggregation_algorithm=self.params.aggregation_algorithm,
                worker_weights=worker_weights,
            ),
            worker_weights,
        )

    def get_pool_input_objects(self, pool_id: str) -> List[mapping.Objects]:
        return get_pool_input_objects(self.client, self.task_mapping, pool_id)

    def get_assignments_and_worker_weights(
        self,
        pool_id: str,
        pool_input_objects: Optional[List[mapping.Objects]] = None,
    ) -> Tuple[List[mapping.Objects], List[toloka.Assignment], Optional[Dict[str, float]]]:
        if self.model_ws:
            # TODO(DATAFORGE-75): model can only substitute all human workers
            #  mixture of model and human solutions is not implemented yet
            pool_input_objects = pool_input_objects or get_pool_input_objects(self.client, self.task_mapping, pool_id)
            assignment = self.model_ws.get_solutions(pool_input_objects, toloka.Assignment.ACCEPTED)
            return pool_input_objects, [assignment], {self.model_ws.model.id: 1.0}

        return get_assignments_and_worker_weights(
            self.client,
            self.task_mapping,
            pool_id,
            self.params.aggregation_algorithm,
            self.assignment_evaluation_strategy,
            pool_input_objects,
        )


class MOSLoop(ClassificationLoop):
    def __init__(
        self,
        client: toloka.TolokaClient,
        task_mapping: mapping.TaskMapping,
        params: Params,
        lang: str,
        interrupter: Optional[utils.LaunchInterrupter] = None,
        with_control_tasks: bool = False,
        model: Optional[worker.Model] = None,
        inputs_to_metadata: Optional[Dict[mapping.Objects, mos.ObjectsMetadata]] = None,
    ):
        assert not with_control_tasks
        super().__init__(client, task_mapping, params, lang, interrupter, with_control_tasks, model)
        speed_rules = params.control.filter_rules(
            predicate_type=control.AssignmentDurationPredicate, action_type=control.BlockUser
        )
        assert len(speed_rules) == 1
        predicate: control.AssignmentDurationPredicate = speed_rules[0].predicate
        self.assignment_evaluation_strategy = mos.MOSEvaluationStrategy(
            inputs_to_metadata,
            task_mapping,
            predicate,
            task_duration_function=params.task_duration_function,
        )

    def get_task_id_to_overlap_increase(self, pool_id: str):
        return {}

    def prior_filtration_is_enabled(self) -> bool:
        return False


# in case of classification task, pool input objects are specified at start
# in case of feedback loop, we don't know pool input objects for check in advance
# here order of pool input objects is not important, only in getting results for classification task
def get_pool_input_objects(
    client: toloka.TolokaClient,
    task_mapping: mapping.TaskMapping,
    pool_id: str,
) -> List[mapping.Objects]:
    pool_input_objects = []
    for task in client.get_tasks(pool_id):
        if not task.known_solutions:
            pool_input_objects.append(task_mapping.from_task(task))
    return pool_input_objects


def get_assignments_solutions(
    client: toloka.TolokaClient,
    task_mapping: mapping.TaskMapping,
    pool_id: str,
    status: List[toloka.Assignment.Status],
    with_control_tasks: bool = False,
) -> List[mapping.AssignmentSolutions]:
    return mapping.get_assignments_solutions(
        assignments=list(client.get_assignments(status=status, pool_id=pool_id)),
        mapping=task_mapping,
        with_control_tasks=with_control_tasks,
    )


def calculate_label_probas(
    client: toloka.TolokaClient,
    task_mapping: mapping.TaskMapping,
    assignment_evaluation_strategy: evaluation.AssignmentAccuracyEvaluationStrategy,
    aggregation_algorithm: classification.AggregationAlgorithm,
    assignment_solutions: List[mapping.AssignmentSolutions],
    pool_id: str,
) -> Dict[mapping.TaskID, classification.LabelProba]:
    task_id_to_label_confidence = {}
    worker_weights = evaluation.calculate_worker_weights(assignment_solutions, assignment_evaluation_strategy)
    input_objects = get_pool_input_objects(client, task_mapping, pool_id)
    accepted_assignments = [
        assignment for assignment, _ in assignment_solutions if assignment.status == toloka.Assignment.ACCEPTED
    ]
    tasks_probas = [
        classification.get_most_probable_label(probas)
        for probas, _ in classification.collect_labels_probas_from_assignments(
            assignments=accepted_assignments,
            task_mapping=task_mapping,
            pool_input_objects=input_objects,
            aggregation_algorithm=aggregation_algorithm,
            worker_weights=worker_weights,
        )
    ]
    for task_input_objects, probas in zip(input_objects, tasks_probas):
        if probas is not None:
            task_id = mapping.TaskID(task_input_objects)
            task_id_to_label_confidence[task_id] = probas
    return task_id_to_label_confidence


def rework_not_finalized_tasks(
    client: toloka.TolokaClient,
    assignments: List[mapping.AssignmentSolutions],
    task_mapping: mapping.TaskMapping,
    task_id_to_overlap_increase: Dict[mapping.TaskID, int],
    pool_id: str,
) -> int:
    # full task ID list is available through created tasks, not assignments, because:
    # - input objects can be added iteratively in common case, "pool input objects" notion may be missing
    # - assignments statuses, attempts from which we take into account, can differ depends on scenario
    task_id_map = get_task_id_map(client, pool_id, task_mapping)
    tasks_to_rework = 0
    # model worker solutions do not affect Toloka overlap
    for task_id, attempts in evaluation.get_tasks_attempts(assignments, task_mapping, with_model=False).items():
        overlap_increase = task_id_to_overlap_increase.get(task_id)
        if not overlap_increase:
            continue
        tasks_to_rework += 1
        toloka_task_id = task_id_map[task_id]
        new_overlap = attempts + overlap_increase
        logger.debug(f'rework task "{task_id.id}", set overlap = {new_overlap} for toloka task {toloka_task_id}')
        client.patch_task_overlap_or_min(
            task_id=toloka_task_id, patch=toloka.task.TaskOverlapPatch(overlap=new_overlap)
        )

    if tasks_to_rework > 0:
        client.open_pool(pool_id)

    return tasks_to_rework


def get_task_id_map(
    client: toloka.TolokaClient,
    pool_id: str,
    task_mapping: mapping.TaskMapping,
    with_control_tasks: bool = False,
) -> Dict[mapping.TaskID, str]:
    return {
        mapping.TaskID(task_mapping.from_task(task)): task.id
        for task in client.get_tasks(pool_id)
        if (with_control_tasks or not task.known_solutions)
    }


def get_assignments_and_worker_weights(
    client: toloka.TolokaClient,
    task_mapping: mapping.TaskMapping,
    pool_id: str,
    aggregation_algorithm: classification.AggregationAlgorithm,
    assignment_evaluation_strategy: evaluation.AssignmentAccuracyEvaluationStrategy,
    pool_input_objects: Optional[List[mapping.Objects]] = None,
) -> Tuple[List[mapping.Objects], List[toloka.Assignment], Optional[Dict[str, float]]]:
    pool_input_objects = pool_input_objects or get_pool_input_objects(client, task_mapping, pool_id)

    all_assignments = get_assignments_solutions(
        client, task_mapping, pool_id, [toloka.Assignment.ACCEPTED, toloka.Assignment.REJECTED], with_control_tasks=True
    )
    accepted_assignments = [
        assignment for assignment, _ in all_assignments if assignment.status == toloka.Assignment.ACCEPTED
    ]
    worker_weights = (
        evaluation.calculate_worker_weights(all_assignments, assignment_evaluation_strategy)
        if aggregation_algorithm == classification.AggregationAlgorithm.MAX_LIKELIHOOD
        else None
    )

    return pool_input_objects, accepted_assignments, worker_weights



import logging
from typing import List, Union, Optional

import toloka.client as toloka

from .. import experts, mapping, pricing, objects
from .params import Params
from .task import PreparedTaskSpec, check_project

logger = logging.getLogger(__name__)


def get_training_name(task_spec: PreparedTaskSpec) -> str:
    return f'Training for {task_spec.log_message()}'


def _log_no_trainings(project_id: str, task_spec: PreparedTaskSpec):
    logger.debug(f'No available trainings found for {task_spec.log_message()} in Toloka project {project_id}')


def _outdated_training_msg(training_id: str, task_spec: PreparedTaskSpec, project_id: str) -> str:
    return f'Found outdated training {training_id} for {task_spec.log_message()} in Toloka project {project_id}'


# TODO (DATAFORGE-58): reliable work with training using lzy
def create_training(
    task_spec: PreparedTaskSpec,
    solutions: List[Union[mapping.TaskSingleSolution, mapping.TaskMultipleSolutionOptions]],
    comments: List[objects.Text],
    client: toloka.TolokaClient,
    training_config: pricing.TrainingConfig,
    append_to_existing: bool = False,
):
    prj = check_project(task_spec, client)

    found_training = experts.find_training(prj.id, client)
    is_ok = found_training is not None and experts.training_is_suitable(
        found_training, training_config.get_task_duration_hint()
    )

    if found_training is None or not append_to_existing:
        _log_no_trainings(prj.id, task_spec)
        training = experts.create_training(prj.id, get_training_name(task_spec), training_config, client)
        logger.debug(f'Training {training.id} created')
    elif is_ok:
        training = found_training
        logger.debug(f'Reusing suitable training {training.id}')
    else:
        existing_training = found_training
        logger.debug(
            f'{_outdated_training_msg(existing_training.id, task_spec, prj.id)}, creating a new one with task reusing'
        )
        training = experts.create_training(
            prj.id, get_training_name(task_spec), training_config, client, reuse_tasks_from=existing_training
        )
        logger.debug(f'Training {training.id} created, reused tasks from {existing_training.id}')

    experts.add_training_tasks(training.id, task_spec.task_mapping, solutions, comments, client)
    client.open_training(training.id)


# TODO (DATAFORGE-58): reliable work with training using lzy
def find_training_requirement(
    project_id: str,
    task_spec: PreparedTaskSpec,
    client: toloka.TolokaClient,
    params: Params,
) -> Optional[toloka.pool.QualityControl.TrainingRequirement]:
    if params.worker_filter.training_score is None:
        return None
    found_training = experts.find_training(project_id, client)
    if found_training is None:
        _log_no_trainings(project_id, task_spec)
        return None
    if not experts.training_is_suitable(found_training, params.task_duration_hint):
        outdated_training_msg = _outdated_training_msg(found_training.id, task_spec, project_id)
        available_task_count = len(list(client.get_tasks(pool_id=found_training.id)))
        try:
            training_config = pricing.TrainingConfig.from_toloka_training(
                found_training, params.task_duration_hint, available_task_count
            )
        except AssertionError as e:
            # task duration hint changes significantly, no way to correctly recreate training
            logger.warn(
                f'{outdated_training_msg}. Failed to recreate training due to a change in task duration hint, '
                f'using outdated version. Consider the following problem: "{e}".'
            )
        else:
            logger.debug(f'{outdated_training_msg}. Recreating with task reusing.')
            found_training = experts.create_training(
                project_id, get_training_name(task_spec), training_config, client, reuse_tasks_from=found_training
            )
            client.open_training(found_training.id)

    logger.debug(f'Using training {found_training.id}')

    return toloka.pool.QualityControl.TrainingRequirement(
        training_pool_id=found_training.id,
        training_passing_skill_value=params.worker_filter.training_score,
    )



from collections import defaultdict
import enum
from typing import List, Tuple, Dict, Optional, Type

import crowdkit.aggregation as aggregation
import pandas as pd
import toloka.client as toloka

from .. import base, mapping, objects, Worker, Human


class AggregationAlgorithm(enum.Enum):
    MAJORITY_VOTE = 'majority-vote'
    MAX_LIKELIHOOD = 'max-likelihood'
    DAWID_SKENE = 'dawid-skene'


WorkerLabel = Tuple[base.Label, Worker]
TaskLabels = List[WorkerLabel]
LabelProba = Tuple[base.Label, float]
TaskLabelsProbas = Dict[base.Label, float]

aggregator_map = {
    AggregationAlgorithm.MAJORITY_VOTE: aggregation.MajorityVote(),
    AggregationAlgorithm.DAWID_SKENE: aggregation.DawidSkene(n_iter=10),
}


def split_generated_answer(combined_type: Type[objects.CombinedAnswer]) -> List[List[objects.CombinedAnswer]]:
    answer_groups = defaultdict(list)
    for instance in combined_type.possible_instances():
        answer_groups[type(instance.get_original_answer())].append(instance)

    return list(answer_groups.values())


def predict_labels_probas(
    labels: List[TaskLabels],
    aggregation_algorithm: AggregationAlgorithm,
    task_mapping: mapping.TaskMapping,
    worker_weights: Optional[Dict[str, float]] = None,
) -> List[Optional[TaskLabelsProbas]]:
    label_cls = task_mapping.output_mapping[0].obj_meta.type
    assert issubclass(label_cls, base.Label)
    if not issubclass(label_cls, objects.CombinedAnswer):
        classes = label_cls.possible_instances()
        extra_data = None if worker_weights is None else (worker_weights, classes)
        return predict_labels_probas_for_class(labels, aggregation_algorithm, extra_data)

    results = [None for _ in labels]

    for classes in split_generated_answer(label_cls):
        indices, current_labels = [], []
        for i, task_labels in enumerate(labels):
            if not task_labels:
                continue
            matches = sum(label in classes for label, _ in task_labels)
            assert matches == 0 or matches == len(task_labels), 'Found mixed-class output labels'

            if matches:
                current_labels.append(task_labels)
                indices.append(i)

        extra_data = None if worker_weights is None else (worker_weights, classes)
        label_probas = predict_labels_probas_for_class(current_labels, aggregation_algorithm, extra_data)
        for i, probas in zip(indices, label_probas):
            results[i] = probas

    return results


def predict_labels_probas_for_class(
    labels: List[TaskLabels],
    aggregation_algorithm: AggregationAlgorithm,
    worker_weights_and_classes: Optional[Tuple[Dict[str, float], List[base.Label]]] = None,
) -> List[Optional[TaskLabelsProbas]]:
    if not labels:
        return []
    if aggregation_algorithm in aggregator_map:
        crowdkit_labels = []
        for task_idx, task_labels in enumerate(labels):
            for label, worker in task_labels:
                crowdkit_labels.append({'label': label, 'task': task_idx, 'worker': worker.id})
        aggregator = aggregator_map[aggregation_algorithm]
        results = [None for _ in labels]

        if not crowdkit_labels:
            return results
        crowdkit_label_probas = aggregator.fit_predict_proba(pd.DataFrame(crowdkit_labels))
        for task_idx, probas in zip(crowdkit_label_probas.index, crowdkit_label_probas.values):
            label_probas = {}
            for label_idx, proba in enumerate(probas):
                label = crowdkit_label_probas.columns[label_idx]
                label_probas[label] = proba
            results[task_idx] = label_probas
        return results

    assert aggregation_algorithm == AggregationAlgorithm.MAX_LIKELIHOOD
    assert worker_weights_and_classes is not None
    worker_weights, classes = worker_weights_and_classes
    num_classes = len(classes)

    def bayes_proba(answer: base.Label, answers: TaskLabels) -> float:
        result = 1.0
        for other_answer, w in answers:
            q = worker_weights[w.id]
            if other_answer == answer:
                result *= q
            else:
                result *= (1 - q) / (num_classes - 1)
        return result

    weighted_results = []
    for task_labels in labels:
        if not task_labels:
            weighted_results.append(None)
            continue
        calculated_z_proba = {label: bayes_proba(label, task_labels) * (1 / num_classes) for label in classes}
        normalization = sum(calculated_z_proba.values())
        weighted_results.append({label: value / normalization for label, value in calculated_z_proba.items()})

    return weighted_results


def get_most_probable_label(label_probas: Optional[TaskLabelsProbas]) -> Optional[LabelProba]:
    if label_probas is None:
        return None
    return max(label_probas.items(), key=lambda label_proba: label_proba[1])


def collect_labels_probas_from_assignments(
    assignments: List[toloka.Assignment],
    task_mapping: mapping.TaskMapping,
    pool_input_objects: List[mapping.Objects],
    aggregation_algorithm: AggregationAlgorithm,
    worker_weights: Optional[Dict[str, float]] = None,
) -> List[Tuple[Optional[TaskLabelsProbas], List[WorkerLabel]]]:
    raw_labels = []
    for _, output_objects_list in mapping.get_solutions(assignments, task_mapping, pool_input_objects):
        task_labels = []
        for output_objects, assignment in output_objects_list:
            # for now label is always the first output object; composite labels not supported yet
            label = output_objects[0]
            assert isinstance(label, base.Label)
            worker = Human(assignment)  # TODO(DATAFORGE-75): can be model
            task_labels.append((label, worker))
        raw_labels.append(task_labels)

    return list(zip(predict_labels_probas(raw_labels, aggregation_algorithm, task_mapping, worker_weights), raw_labels))



css = """
<style>

    body {
        scroll-behavior: smooth;
    }

    h2, h3, h4 {
        font-weight: inherit;
        font-style: four;
        text-align: left;
        text-transform: none;
        margin: 20px 0 10px 0;
    }

    p {
        font-size: 14px;
        align: left;
        color: #000000;
        margin: 2px 0;
        padding-right: 10px;
    }

    ul, ol {
        padding-left: 15px;
        margin-left: 25px;
    }

    li {
        font-size: 14px;
        align: left;
        color: #000000;
        margin: 2px 0;
        padding-right: 10px;
    }

    .hide {
        display: none
    }

    .hide + label ~ div {
        display: none
    }

    .hide + label ~ p {
        display: none
    }

    .hide + label {
        border-bottom: 1px dotted blue;
        padding: 0;
        margin: 0;
        cursor: pointer;
        display: inline-block;
        color: blue;
        font-size: 14px;
    }

    .hide:checked + label {
        border-bottom: 0;
        color: green;
        padding-right: 10px;
        margin: 0;
        font-size: 14px;
    }

    .hide:checked + label + div {
        display: block;
    }

    .hide:checked + label + span {
        display: block;
    }

    .hide:checked + label + p {
        display: block;
    }

    img {
        width: 100%;
        max-width: 600px;
    }

</style>
"""



from collections import Counter
import datetime
from typing import List, Dict, Optional, Tuple

import toloka.client as toloka
import toloka.client.project.template_builder as tb
import uuid


CODE_LENGTH = 10
PROJECT_PRIVATE_COMMENT = 'expert-registration'
EXPERT_SKILL = 'expert'


def generate_code() -> str:
    return uuid.uuid4().hex[:10]


def generate_codes(worker_names: List[str]) -> Dict[str, str]:
    return {name: generate_code() for name in worker_names}


def get_skill_name(task_id: Optional[str], lang: Optional[str]) -> str:
    return f"{EXPERT_SKILL}{'_' + task_id if task_id is not None else ''}{'_' + lang if lang is not None else ''}"


def get_skill_names(task_id: Optional[str], lang: Optional[str]) -> List[str]:
    skill_names = []
    task_ids = {task_id, None}
    langs = {lang, None}
    for task_id in task_ids:
        for lang in langs:
            skill_names.append(get_skill_name(task_id, lang))
    return sorted(skill_names)


def get_skills(
    task_id: Optional[str],
    lang: Optional[str],
    client: toloka.TolokaClient,
) -> List[Tuple[str, toloka.Skill]]:
    skill_names = get_skill_names(task_id, lang)
    skills = [next(client.get_skills(name=skill_name), None) for skill_name in skill_names]
    return [(name, skill) for name, skill in zip(skill_names, skills) if skill is not None]


def create_registration_project(client: toloka.TolokaClient) -> toloka.Project:
    project = toloka.Project(
        public_name=f"Welcome to {client.get_requester().public_name['EN']}!",
        public_description='Welcome! You will need to enter a code',
        public_instructions='Enter supplied code',
        private_comment=PROJECT_PRIVATE_COMMENT,
        task_spec=toloka.project.task_spec.TaskSpec(
            input_spec={'unused': toloka.project.field_spec.StringSpec(required=False, hidden=True)},
            output_spec={
                'code': toloka.project.field_spec.StringSpec(
                    required=True, min_length=CODE_LENGTH, max_length=CODE_LENGTH
                )
            },
            view_spec=toloka.project.TemplateBuilderViewSpec(
                view=tb.ListViewV1(
                    items=[
                        tb.TextViewV1(content='Enter code:'),
                        tb.TextFieldV1(
                            data=tb.OutputData(path='code'),
                            validation=tb.SchemaConditionV1(
                                schema={
                                    'type': 'string',
                                    'pattern': f'[a-zA-Z0-9]{{{CODE_LENGTH}}}',
                                    'minLength': CODE_LENGTH,
                                    'maxLength': CODE_LENGTH,
                                },
                                hint=f'Code length should be {CODE_LENGTH}, '
                                'you can use only latin alphabet and digits',
                            ),
                        ),
                    ]
                )
            ),
        ),
    )

    return client.create_project(project)


def create_registration_pool(
    client: toloka.TolokaClient,
    expected_date: str = '14.05.1405',
    task_id: Optional[str] = None,
    lang: Optional[str] = None,
) -> toloka.Pool:
    project_id = None
    for project in client.get_projects(status='ACTIVE'):
        if project.private_comment == PROJECT_PRIVATE_COMMENT:
            project_id = project.id
            break
    if project_id is None:
        project_id = create_registration_project(client).id
    skill_name = get_skill_name(task_id, lang)
    expert_skill = next(client.get_skills(name=skill_name), None)
    if expert_skill is None:
        expert_skill = client.create_skill(name=skill_name, hidden=True)
    date = datetime.datetime.strptime(expected_date, '%d.%m.%Y')
    for pool in client.get_pools(project_id=project_id, status='OPEN'):
        if pool.private_comment == skill_name:
            return pool
    pool = toloka.Pool(
        project_id=project_id,
        public_description=skill_name,
        private_comment=skill_name,
        private_name='Pool for expert promotion',
        auto_accept_solutions=False,
        may_contain_adult_content=False,
        reward_per_assignment=0.0,
        assignment_max_duration_seconds=120,
        will_expire=datetime.datetime.utcnow() + datetime.timedelta(days=365),
        filter=(
            (toloka.filter.Skill(expert_skill.id) == None)  # noqa
            & (toloka.filter.DateOfBirth < int((date + datetime.timedelta(days=2)).timestamp()))
            & (toloka.filter.DateOfBirth > int((date - datetime.timedelta(days=2)).timestamp()))
        ),
    )
    pool.set_mixer_config(golden_tasks_count=1)

    pool.quality_control.add_action(
        collector=toloka.collectors.AnswerCount(),
        conditions=[toloka.conditions.AssignmentsAcceptedCount > 0],
        action=toloka.actions.SetSkill(skill_id=expert_skill.id, skill_value=1),
    )

    pool = client.create_pool(pool)
    client.create_task(
        toloka.task.Task(
            input_values={'unused': ''},
            known_solutions=[toloka.task.BaseTask.KnownSolution(output_values={'code': 'AAAAAAAAAA'})],
            pool_id=pool.id,
            infinite_overlap=True,
        )
    )
    client.open_pool(pool.id)
    return pool


def get_entered_code(assignment: toloka.Assignment) -> str:
    return assignment.solutions[0].output_values['code']


def approve_expected_codes(client: toloka.TolokaClient, pool_id: str, name_to_code: Dict[str, str]) -> Dict[str, str]:
    name_to_worker_id = {}
    assignments = list(client.get_assignments(pool_id=pool_id, status=[toloka.Assignment.SUBMITTED]))
    code_counter = Counter(get_entered_code(assignment) for assignment in assignments)
    code_to_name = {v: k for k, v in name_to_code.items()}
    for assignment in assignments:
        worker_id = assignment.user_id
        code = get_entered_code(assignment)
        if code not in code_to_name:
            print(f'Unexpected code {code} from worker {worker_id}')
            client.reject_assignment(assignment_id=assignment.id, public_comment='Unexpected code')
        elif code_counter[code] > 1:
            print(f'More than one answer for code {code}, this one from {worker_id}')
            client.reject_assignment(assignment_id=assignment.id, public_comment='Duplicated code')
        else:
            name_to_worker_id[code_to_name[code]] = worker_id
            client.accept_assignment(assignment_id=assignment.id, public_comment='')
    return name_to_worker_id


def revoke_expert_access(
    client: toloka.TolokaClient, worker_id: str, task_id: Optional[str] = None, lang: Optional[str] = None
):
    skill_name = get_skill_name(task_id, lang)
    skill = next(client.get_skills(name=skill_name), None)
    if skill is None:
        print(f'Warning: no such skill found: {skill_name}')
    else:
        found_skills = list(client.get_user_skills(user_id=worker_id, skill_id=skill.id))
        if not found_skills:
            print(f'Warning: no {skill_name} skill found for worker {worker_id}')
        else:
            for skill in found_skills:
                client.delete_user_skill(skill.id)



import functools
import logging
from typing import List, Dict, Any, Union, Tuple

import requests
from requests.adapters import Retry, HTTPAdapter
from urllib3.exceptions import InsecureRequestWarning

from toloka.client import TolokaClient
from toloka.client.primitives.retry import TolokaRetry
from toloka.client.exceptions import raise_on_api_error

TST_YANDEX_COM = 'https://autotest-external-back.admin.toloka.tst.yandex.com'

requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)

logger = logging.getLogger(__name__)


def _make_user_info(toloka_user: Dict[str, Any]) -> Dict[str, Any]:
    return {
        'type': 'INDIVIDUAL',
        'firstName': 'Testing',
        'lastName': 'Requester',
        'displayName': {'EN': toloka_user['displayName']},
        # 'displayName': toloka_user['displayName'],
        'registrationAddress': {'country': 'BY', 'cityId': 157, 'street': 'street', 'house': 'house', 'index': 'index'},
        'postAddress': {'country': 'BY', 'cityId': 157, 'street': 'street', 'house': 'house', 'index': 'index'},
        'contactPhone': '+375199999999',
        'userLang': 'RU',
        "rsaAccepted": True,
        "subscriptionOnNewsAccepted": False,
        'defaultEmail': toloka_user['login'] + '@yandex.ru'
        # f'{}requester_emaiil@example.org',
    }


def _make_worker_info(toloka_user: Dict[str, Any]) -> Dict[str, Any]:
    # ./autotest/xspec/src/main/groovy/code/worker/WorkerForm.groovy
    return {
        'uid': toloka_user['uid'],
        'login': toloka_user['login'],
        'defaultEmail': toloka_user['login'] + '@yandex.ru',
        'displayName': {'EN': toloka_user['displayName']},
        'citizenship': 'RU',
        'education': 'MIDDLE',
        'userLang': 'RU',
        'fullName': 'Иван Иванов',
        'firstName': 'Иван',
        'lastName': 'Иванов',
        'gender': 'MALE',
        'birthDay': '1990-01-01',
        'country': 'BY',
        'cityId': 157,
        'languages': ['RU', 'BE', 'EN'],
        'adultAllowed': True,
    }


class NDARegisterClient:
    def __init__(self):
        self._session = requests.Session()

        retries = Retry(
            total=20,
            backoff_factor=2,
            status_forcelist=[500, 502, 503, 504],
            allowed_methods=frozenset({'DELETE', 'GET', 'HEAD', 'OPTIONS', 'PUT', 'TRACE', 'POST', 'PATCH'}),
        )
        self._session.mount('https://', HTTPAdapter(max_retries=retries))

    def _create_user(self, user_type: str, url: str = TST_YANDEX_COM) -> Dict[str, Any]:
        response = self._session.post(
            f'{url}/public/testing/env/register', json={'loginSuffix': user_type}, verify=False
        )
        response.raise_for_status()
        logger.debug(f'Created user with oauth {response.json()["oauth"]}')
        return response.json()

    def create_admin(self, url: str = TST_YANDEX_COM) -> Dict[str, Any]:
        toloka_user = self._create_user('adm', url=url)
        # Grant admin privileges
        response = self._session.post(
            f'{url}/api/dmz/dev/admin', json={'uid': toloka_user['uid'], 'allowedToSpy': True}, verify=False
        )
        response.raise_for_status()
        return toloka_user

    def create_user(self, user_type: str = 'worker', url: str = TST_YANDEX_COM) -> Dict[str, Any]:
        registration_type = {
            'worker': 'wrk',
            'requester': 'req',
        }[user_type]
        user = self._create_user(registration_type)
        headers = {'Authorization': f'OAuth {user["oauth"]}'}

        # Making an attempt to set data to get an error with a conformation key
        user_info = {'worker': _make_worker_info, 'requester': _make_user_info}[user_type](user)

        if user_type == 'worker':
            user_info['displayName'] = user_info['displayName']['EN']
        response = self._session.post(
            f'{url}/api/users/current/{user_type}', headers=headers, json=user_info, verify=False
        )
        assert response.status_code == 409, (response.status_code, response.json())
        assert response.json()['code'] == 'NEED_PHONE_CONFIRMATION', response.json()
        key = response.json()['payload']['key']

        # Confirming data as if by SMS
        code = self._session.get(f'{url}/public/testing/env/sms?trackId={key}', verify=False).json()
        response = self._session.post(
            f'{url}/api/passport/sms/validate', headers=headers, verify=False, json={'key': key, 'code': code}
        )
        response.raise_for_status()

        # Actually setting user data
        response = self._session.post(
            f'{url}/api/users/current/{user_type}', headers=headers, json=user_info, verify=False
        )
        response.raise_for_status()
        return {**response.json(), **user}


class NDATolokaClient(TolokaClient):
    def __init__(self, token: str):
        retries = 60
        retry_quotas = TolokaRetry.Unit.MIN

        def retryer_factory(
            retries: int,
            retry_quotas: Union[List[str], str, None],
            status_list: Tuple[int] = (408, 409, 429, 500, 503, 504),
        ):
            return TolokaRetry(
                retry_quotas=retry_quotas,
                total=retries,
                status_forcelist=list(status_list),
                allowed_methods=['HEAD', 'GET', 'PUT', 'DELETE', 'OPTIONS', 'TRACE', 'POST', 'PATCH'],
                backoff_factor=2,  # summary retry time more than 10 seconds
            )

        super().__init__(
            token,
            url=TST_YANDEX_COM,
            retryer_factory=functools.partial(retryer_factory, retries, retry_quotas),
            timeout=(60.0, 240.0),
        )

    def _raw_request(self, method: str, path: str, **kwargs: Any) -> requests.Response:
        # return super()._raw_request(method, path, verify=False, **kwargs)

        if kwargs.get('params'):
            params = kwargs['params']
            for key, value in params.items():
                if isinstance(value, bool):
                    params[key] = 'true' if value else 'false'
        if self.default_timeout is not None and 'timeout' not in kwargs:
            kwargs['timeout'] = self.default_timeout

        response = self._session.request(method, f'{self.url}/api{path}', verify=False, **kwargs)
        raise_on_api_error(response)
        return response

    # For worker
    # AssignmentService.groovy
    def request_assignment(self, pool_id: str) -> Dict[str, Any]:
        return self._request('post', '/i-v2/assignment-executions', json={'poolId': pool_id})

    def skip_assignment(self, assignment_id: str) -> Dict[str, Any]:
        return self._request('post', f'/i-v2/assignment-executions/{assignment_id}/skip')

    def expire_assignment(self, assignment_id: str) -> Dict[str, Any]:
        return self._request('post', f'/i-v2/assignment-executions/{assignment_id}/expire')

    def submit_assignment(self, assignment_id: str, solutions: List[Dict[str, Any]]) -> Dict[str, Any]:
        solutions = {'solutions': [{'output_values': solution} for solution in solutions]}
        return self._request('post', f'/i-v2/assignment-executions/{assignment_id}/submit', json=solutions)

    # Requires admin privileges
    def change_credit(self, user_id: str, credit_diff: int) -> Dict[str, Any]:
        return self._request('post', f'/admin/requesters/{user_id}/change-credit', json={'creditDiff': credit_diff})

    def resolve_id(self, uid: str) -> Dict[str, Any]:
        return self._request('get', f'/dmz/dev/resolve-worker-id/{uid}')


class CallRecorderClient(NDATolokaClient):
    def __init__(self, *args, **kwargs):
        self.calls: List[Tuple[str, str, Tuple[Any], Dict[str, Any], Dict[str, Any]]] = []
        super().__init__(*args, **kwargs)

    def _raw_request(self, method: str, path: str, *args, **kwargs):
        result = super()._raw_request(method, path, *args, **kwargs)
        self.calls.append((method, path, args, kwargs, result.json()))
        return result



from dataclasses import dataclass
from typing import Type, TypeVar, Optional

from dataforge.base import Object, ObjectMeta, TextFormat, LocalizedString


@dataclass(frozen=True)
class Text(Object):
    text: str


TextT = TypeVar('TextT', bound=Type[Text], covariant=True)


@dataclass
class TextValidation:
    regex: LocalizedString
    hint: LocalizedString

    def __post_init__(self):
        LocalizedString.convert_dataclass(self)


@dataclass
class TextMeta(ObjectMeta[TextT]):
    format: TextFormat = TextFormat.PLAIN
    validation: Optional[TextValidation] = None


@dataclass(frozen=True)
class Audio(Object):
    url: str

    @staticmethod
    def is_media() -> bool:
        return True


@dataclass(frozen=True)
class Image(Object):
    url: str

    @staticmethod
    def is_media() -> bool:
        return True


@dataclass(frozen=True)
class Video(Object):
    url: str

    @staticmethod
    def is_media() -> bool:
        return True



import abc
from dataclasses import dataclass
import datetime
from typing import Optional

import toloka.client as toloka

from . import control, worker


class PoolBuilder:
    def __init__(
        self,
        project_id: str,
        private_name: str,
        public_description: Optional[str],
        reward_per_assignment: float,
        assignment_duration_hint: datetime.timedelta,
        real_tasks_count: int,
        control_tasks_count: int,
        ttl_days: int = 30,
        auto_accept_period_day: int = 14,
        priority: int = 30,
        may_contain_adult_content=True,
    ):
        pool = toloka.Pool(
            project_id=project_id,
            private_name=private_name,
            public_description=public_description,
            may_contain_adult_content=may_contain_adult_content,
            will_expire=datetime.datetime.utcnow() + datetime.timedelta(days=ttl_days),
            reward_per_assignment=reward_per_assignment,
            assignment_max_duration_seconds=int((assignment_duration_hint * 5).total_seconds()),
            auto_accept_solutions=False,
            auto_accept_period_day=auto_accept_period_day,
            assignments_issuing_config=toloka.Pool.AssignmentsIssuingConfig(issue_task_suites_in_creation_order=False),
            priority=priority,
        )

        pool.set_mixer_config(
            real_tasks_count=real_tasks_count, golden_tasks_count=control_tasks_count, training_tasks_count=0
        )

        self.pool = pool
        self.real_tasks_count = real_tasks_count
        self.control_tasks_count = control_tasks_count
        self.tasks_in_assignment = real_tasks_count + control_tasks_count
        self.may_contain_adult_content = may_contain_adult_content
        self.assignment_duration_hint = assignment_duration_hint

    def build(self) -> toloka.Pool:
        return self.pool

    def set_overlap(self, overlap: int) -> 'PoolBuilder':
        self.pool.set_defaults(
            defaults=toloka.Pool.Defaults(
                default_overlap_for_new_tasks=overlap, default_overlap_for_new_task_suites=overlap
            )
        )
        return self

    def set_worker_filter(self, human_filter: Optional[worker.HumanFilter]) -> 'PoolBuilder':
        if human_filter is None:
            return self
        assert self.pool.filter is None, 'repeated set of worker filter'
        self.pool.filter = human_filter.to_toloka_filter()
        return self

    def add_captcha(
        self,
        frequency: toloka.pool.QualityControl.CaptchaFrequency = toloka.pool.QualityControl.CaptchaFrequency.LOW,
        history_size: int = 10,
        min_completed_count: int = 5,
        min_success_rate: float = 30.0,
        duration: int = 8,
        duration_unit: toloka.user_restriction.DurationUnit = toloka.user_restriction.DurationUnit.HOURS,
        scope: toloka.UserRestriction = toloka.UserRestriction.PROJECT,
    ) -> 'PoolBuilder':
        self.pool.set_captcha_frequency(frequency)
        self.pool.quality_control.add_action(
            collector=toloka.collectors.Captcha(history_size=history_size),
            conditions=[
                toloka.conditions.StoredResultsCount >= min_completed_count,
                toloka.conditions.SuccessRate <= min_success_rate,
            ],
            action=toloka.actions.RestrictionV2(
                private_comment='Captcha', duration=duration, duration_unit=duration_unit, scope=scope
            ),
        )
        return self

    def add_skipped_assignments(
        self,
        skipped_count: int = 5,
        duration: int = 8,
        duration_unit: toloka.user_restriction.DurationUnit = toloka.user_restriction.DurationUnit.HOURS,
        scope: toloka.UserRestriction = toloka.UserRestriction.PROJECT,
    ) -> 'PoolBuilder':
        self.pool.quality_control.add_action(
            collector=toloka.collectors.SkippedInRowAssignments(),
            conditions=[toloka.conditions.SkippedInRowCount >= skipped_count],
            action=toloka.actions.RestrictionV2(
                private_comment='Skipped assignments', duration=duration, duration_unit=duration_unit, scope=scope
            ),
        )
        return self

    def add_vacation(
        self,
        work_duration_before_vacation: datetime.timedelta,
        duration: int = 5,
        duration_unit: toloka.user_restriction.DurationUnit = toloka.user_restriction.DurationUnit.HOURS,
        scope: toloka.UserRestriction = toloka.UserRestriction.PROJECT,
    ) -> 'PoolBuilder':
        # 1h of tasks approx
        accepted_count = int(work_duration_before_vacation / self.assignment_duration_hint)
        assert accepted_count > 0
        self.pool.quality_control.add_action(
            collector=toloka.collectors.AnswerCount(),
            conditions=[toloka.conditions.AssignmentsAcceptedCount >= accepted_count],
            action=toloka.actions.RestrictionV2(
                private_comment='Completed many assignments, vacation',
                duration=duration,
                duration_unit=duration_unit,
                scope=scope,
            ),
        )
        return self

    def add_training(
        self,
        training_requirement: Optional[toloka.pool.QualityControl.TrainingRequirement],
    ) -> 'PoolBuilder':
        if training_requirement is not None:
            self.pool.set_training_requirement(training_requirement)
        return self

    def add_control_rules(self, control_params: Optional[control.Control]) -> 'PoolBuilder':
        if control_params is not None:
            toloka_rules = control_params.to_toloka_quality_control(
                self.control_tasks_count, self.assignment_duration_hint
            )
            for toloka_rule in toloka_rules:
                self.pool.quality_control.add_action(**toloka_rule)
        return self

    def set_auto_accept(self) -> 'PoolBuilder':
        self.pool.auto_accept_solutions = True
        return self

    def add_quality_adjustment(self, percent: Optional[int] = 90):
        if percent:
            self.pool.set_speed_quality_balance(
                toloka.pool.speed_quality_balance_config.TopPercentageByQuality(percent=percent)
            )


@dataclass
class BaseConfig:
    project_id: str
    private_name: str
    reward_per_assignment: float
    task_duration_hint: datetime.timedelta
    real_tasks_count: int
    priority: int = 30
    ttl_days: int = 30
    may_contain_adult_content: bool = True
    public_description: Optional[str] = None

    @abc.abstractmethod
    def get_tasks_count(self) -> int:
        ...


@dataclass
class CrowdConfig(BaseConfig, abc.ABC):
    worker_filter: Optional[worker.BaseWorkerFilter] = None
    training_requirement: Optional[toloka.pool.QualityControl.TrainingRequirement] = None
    auto_accept_period_day: int = 14
    work_duration_before_vacation: datetime.timedelta = datetime.timedelta(hours=1)
    control_params: Optional[control.Control] = None
    adjust_percent: Optional[int] = 90


@dataclass
class MarkupConfig(CrowdConfig):
    # TODO(DATAFORGE-75): works only for current markup model worker limitations
    first_attempt_by_model_worker: bool = False

    def get_tasks_count(self) -> int:
        return self.real_tasks_count


@dataclass
class ClassificationConfig(CrowdConfig):
    control_tasks_count: int = 3
    overlap: int = 3

    def get_tasks_count(self) -> int:
        return self.real_tasks_count + self.control_tasks_count


@dataclass
class ExpertConfig(BaseConfig):
    worker_filter: worker.ExpertFilter = worker.ExpertFilter(skills=[])

    def get_tasks_count(self) -> int:
        return self.real_tasks_count


def create_pool_params(config: CrowdConfig):
    assert isinstance(config, MarkupConfig) or isinstance(config, ClassificationConfig)

    if isinstance(config, MarkupConfig):
        control_tasks_count = 0
        overlap = 0 if config.first_attempt_by_model_worker else 1
    else:
        control_tasks_count = config.control_tasks_count
        overlap = config.overlap

    task_count = config.get_tasks_count()

    assignment_duration_hint = config.task_duration_hint * task_count

    builder = PoolBuilder(
        project_id=config.project_id,
        private_name=config.private_name,
        public_description=config.public_description,
        reward_per_assignment=config.reward_per_assignment,
        assignment_duration_hint=assignment_duration_hint,
        real_tasks_count=config.real_tasks_count,
        control_tasks_count=control_tasks_count,
        priority=config.priority,
        auto_accept_period_day=config.auto_accept_period_day,
        may_contain_adult_content=config.may_contain_adult_content,
        ttl_days=config.ttl_days,
    )

    builder.set_worker_filter(config.worker_filter).add_training(
        config.training_requirement
    ).add_captcha().add_skipped_assignments().add_control_rules(config.control_params).add_vacation(
        work_duration_before_vacation=config.work_duration_before_vacation
    ).set_overlap(
        overlap
    ).add_quality_adjustment(
        config.adjust_percent
    )

    return builder.build()


def create_expert_pool_params(config: ExpertConfig):
    control_tasks_count, overlap = 0, 1

    task_count = config.get_tasks_count()

    assignment_duration_hint = config.task_duration_hint * task_count

    builder = PoolBuilder(
        project_id=config.project_id,
        private_name=config.private_name,
        public_description=config.public_description,
        reward_per_assignment=config.reward_per_assignment,
        assignment_duration_hint=assignment_duration_hint,
        real_tasks_count=config.real_tasks_count,
        control_tasks_count=control_tasks_count,
        priority=config.priority,
        may_contain_adult_content=config.may_contain_adult_content,
        ttl_days=config.ttl_days,
    )

    builder.set_worker_filter(config.worker_filter).set_overlap(overlap).set_auto_accept()

    return builder.build()
